"""
Manage the shadow file on Linux systems

.. important::
    If you feel that Salt should be using this module to manage passwords on a
    minion, and it is using a different module (or gives an error similar to
    *'shadow.info' is not available*), see :ref:`here
    <module-provider-override>`.
"""

import datetime
import functools
import logging
import os

import salt.utils.data
import salt.utils.files
from salt.exceptions import CommandExecutionError

try:
    import spwd
except ImportError:
    pass


try:
    import salt.utils.pycrypto

    HAS_CRYPT = True
except ImportError:
    HAS_CRYPT = False

__virtualname__ = "shadow"

log = logging.getLogger(__name__)


def __virtual__():
    return __virtualname__ if __grains__.get("kernel", "") == "Linux" else False


def default_hash():
    """
    Returns the default hash used for unset passwords

    CLI Example:

    .. code-block:: bash

        salt '*' shadow.default_hash
    """
    return "!"


def info(name, root=None):
    """
    Return information for the specified user

    name
        User to get the information for

    root
        Directory to chroot into

    CLI Example:

    .. code-block:: bash

        salt '*' shadow.info root
    """
    if root is not None:
        getspnam = functools.partial(_getspnam, root=root)
    else:
        getspnam = functools.partial(spwd.getspnam)

    try:
        data = getspnam(name)
        ret = {
            "name": data.sp_namp if hasattr(data, "sp_namp") else data.sp_nam,
            "passwd": data.sp_pwdp if hasattr(data, "sp_pwdp") else data.sp_pwd,
            "lstchg": data.sp_lstchg,
            "min": data.sp_min,
            "max": data.sp_max,
            "warn": data.sp_warn,
            "inact": data.sp_inact,
            "expire": data.sp_expire,
        }
    except (KeyError, FileNotFoundError):
        return {
            "name": "",
            "passwd": "",
            "lstchg": "",
            "min": "",
            "max": "",
            "warn": "",
            "inact": "",
            "expire": "",
        }
    return ret


def _set_attrib(name, key, value, param, root=None, validate=True):
    """
    Set a parameter in /etc/shadow
    """
    pre_info = info(name, root=root)

    # If the user is not present or the attribute is already present,
    # we return early
    if not pre_info["name"]:
        return False

    if value == pre_info[key]:
        return True

    cmd = ["chage"]

    if root is not None:
        cmd.extend(("-R", root))

    cmd.extend((param, value, name))

    ret = not __salt__["cmd.retcode"](cmd, python_shell=False)
    if validate:
        ret = info(name, root=root).get(key) == value
    return ret


def set_inactdays(name, inactdays, root=None):
    """
    Set the number of days of inactivity after a password has expired before
    the account is locked. See man chage.

    name
        User to modify

    inactdays
        Set password inactive after this number of days

    root
        Directory to chroot into

    CLI Example:

    .. code-block:: bash

        salt '*' shadow.set_inactdays username 7
    """
    return _set_attrib(name, "inact", inactdays, "-I", root=root)


def set_maxdays(name, maxdays, root=None):
    """
    Set the maximum number of days during which a password is valid.
    See man chage.

    name
        User to modify

    maxdays
        Maximum number of days during which a password is valid

    root
        Directory to chroot into

    CLI Example:

    .. code-block:: bash

        salt '*' shadow.set_maxdays username 90
    """
    return _set_attrib(name, "max", maxdays, "-M", root=root)


def set_mindays(name, mindays, root=None):
    """
    Set the minimum number of days between password changes. See man chage.

    name
        User to modify

    mindays
        Minimum number of days between password changes

    root
        Directory to chroot into

    CLI Example:

    .. code-block:: bash

        salt '*' shadow.set_mindays username 7
    """
    return _set_attrib(name, "min", mindays, "-m", root=root)


def gen_password(password, crypt_salt=None, algorithm="sha512"):
    """
    .. versionadded:: 2014.7.0

    Generate hashed password

    .. note::

        When called this function is called directly via remote-execution,
        the password argument may be displayed in the system's process list.
        This may be a security risk on certain systems.

    password
        Plaintext password to be hashed.

    crypt_salt
        Crpytographic salt. If not given, a random 8-character salt will be
        generated.

    algorithm
        The following hash algorithms are supported:

        * md5
        * blowfish (not in mainline glibc, only available in distros that add it)
        * sha256
        * sha512 (default)

    CLI Example:

    .. code-block:: bash

        salt '*' shadow.gen_password 'I_am_password'
        salt '*' shadow.gen_password 'I_am_password' crypt_salt='I_am_salt' algorithm=sha256
    """
    if not HAS_CRYPT:
        raise CommandExecutionError(
            "gen_password is not available on this operating system "
            'because the "crypt" python module is not available.'
        )
    return salt.utils.pycrypto.gen_hash(crypt_salt, password, algorithm)


def del_password(name, root=None):
    """
    .. versionadded:: 2014.7.0

    Delete the password from name user

    name
        User to delete

    root
        Directory to chroot into

    CLI Example:

    .. code-block:: bash

        salt '*' shadow.del_password username
    """
    cmd = ["passwd"]
    if root is not None:
        cmd.extend(("-R", root))
    cmd.extend(("-d", name))

    __salt__["cmd.run"](cmd, python_shell=False, output_loglevel="quiet")
    uinfo = info(name, root=root)
    return not uinfo["passwd"] and uinfo["name"] == name


def lock_password(name, root=None):
    """
    .. versionadded:: 2016.11.0

    Lock the password from specified user

    name
        User to lock

    root
        Directory to chroot into

    CLI Example:

    .. code-block:: bash

        salt '*' shadow.lock_password username
    """
    pre_info = info(name, root=root)
    if not pre_info["name"]:
        return False

    if pre_info["passwd"].startswith("!"):
        return True

    cmd = ["passwd"]

    if root is not None:
        cmd.extend(("-R", root))

    cmd.extend(("-l", name))

    __salt__["cmd.run"](cmd, python_shell=False)
    return info(name, root=root)["passwd"].startswith("!")


def unlock_password(name, root=None):
    """
    .. versionadded:: 2016.11.0

    Unlock the password from name user

    name
        User to unlock

    root
        Directory to chroot into

    CLI Example:

    .. code-block:: bash

        salt '*' shadow.unlock_password username
    """
    pre_info = info(name, root=root)
    if not pre_info["name"]:
        return False

    if not pre_info["passwd"].startswith("!"):
        return True

    cmd = ["passwd"]

    if root is not None:
        cmd.extend(("-R", root))

    cmd.extend(("-u", name))

    __salt__["cmd.run"](cmd, python_shell=False)
    return not info(name, root=root)["passwd"].startswith("!")


def set_password(name, password, use_usermod=False, root=None):
    """
    Set the password for a named user. The password must be a properly defined
    hash. A password hash can be generated with :py:func:`gen_password`.

    name
        User to set the password

    password
        Password already hashed

    use_usermod
        Use usermod command to better compatibility

    root
        Directory to chroot into

    CLI Example:

    .. code-block:: bash

        salt '*' shadow.set_password root '$1$UYCIxa628.9qXjpQCjM4a..'
    """
    if __salt__["cmd.retcode"](["id", name], ignore_retcode=True) != 0:
        log.warning("user %s does not exist, cannot set password", name)
        return False

    if not salt.utils.data.is_true(use_usermod):
        # Edit the shadow file directly
        # ALT Linux uses tcb to store password hashes. More information found
        # in manpage (http://docs.altlinux.org/manpages/tcb.5.html)
        if __grains__["os"] == "ALT":
            s_file = f"/etc/tcb/{name}/shadow"
        else:
            s_file = "/etc/shadow"
        if root:
            s_file = os.path.join(root, os.path.relpath(s_file, os.path.sep))

        ret = {}
        if not os.path.isfile(s_file):
            return ret
        lines = []
        user_found = False
        lstchg = str((datetime.datetime.today() - datetime.datetime(1970, 1, 1)).days)
        with salt.utils.files.fopen(s_file, "r") as fp_:
            for line in fp_:
                # Fix malformed entry by first ignoring extra fields, then
                # adding missing fields.
                comps = line.strip().split(":")[:9]
                if comps[0] == name:
                    num_missing = 9 - len(comps)
                    if num_missing:
                        comps.extend([""] * num_missing)
                    user_found = True
                    comps[1] = password
                    comps[2] = lstchg
                    line = ":".join(comps) + "\n"
                lines.append(line)
        if not user_found:
            log.warning("shadow entry not present for user %s, adding", name)
            with salt.utils.files.fopen(s_file, "a+") as fp_:
                fp_.write(
                    "{name}:{password}:{lstchg}::::::\n".format(
                        name=name, password=password, lstchg=lstchg
                    )
                )
        else:
            with salt.utils.files.fopen(s_file, "w+") as fp_:
                fp_.writelines(lines)
        uinfo = info(name, root=root)
        return uinfo["passwd"] == password
    else:
        # Use usermod -p (less secure, but more feature-complete)
        cmd = ["usermod"]
        if root is not None:
            cmd.extend(("-R", root))
        cmd.extend(("-p", password, name))

        __salt__["cmd.run"](cmd, python_shell=False, output_loglevel="quiet")
        uinfo = info(name, root=root)
        return uinfo["passwd"] == password


def set_warndays(name, warndays, root=None):
    """
    Set the number of days of warning before a password change is required.
    See man chage.

    name
        User to modify

    warndays
        Number of days of warning before a password change is required

    root
        Directory to chroot into

    CLI Example:

    .. code-block:: bash

        salt '*' shadow.set_warndays username 7
    """
    return _set_attrib(name, "warn", warndays, "-W", root=root)


def set_date(name, date, root=None):
    """
    Sets the value for the date the password was last changed to days since the
    epoch (January 1, 1970). See man chage.

    name
        User to modify

    date
        Date the password was last changed

    root
        Directory to chroot into

    CLI Example:

    .. code-block:: bash

        salt '*' shadow.set_date username 0
    """
    return _set_attrib(name, "lstchg", date, "-d", root=root, validate=False)


def set_expire(name, expire, root=None):
    """
    .. versionchanged:: 2014.7.0

    Sets the value for the date the account expires as days since the epoch
    (January 1, 1970). Using a value of -1 will clear expiration. See man
    chage.

    name
        User to modify

    date
        Date the account expires

    root
        Directory to chroot into

    CLI Example:

    .. code-block:: bash

        salt '*' shadow.set_expire username -1
    """
    return _set_attrib(name, "expire", expire, "-E", root=root, validate=False)


def list_users(root=None):
    """
    .. versionadded:: 2018.3.0

    Return a list of all shadow users

    root
        Directory to chroot into

    CLI Example:

    .. code-block:: bash

        salt '*' shadow.list_users
    """
    if root is not None:
        getspall = functools.partial(_getspall, root=root)
    else:
        getspall = functools.partial(spwd.getspall)

    return sorted(
        user.sp_namp if hasattr(user, "sp_namp") else user.sp_nam for user in getspall()
    )


def _getspnam(name, root=None):
    """
    Alternative implementation for getspnam, that use only /etc/shadow
    """
    root = "/" if not root else root
    passwd = os.path.join(root, "etc/shadow")
    with salt.utils.files.fopen(passwd) as fp_:
        for line in fp_:
            comps = line.strip().split(":")
            if comps[0] == name:
                # Generate a getspnam compatible output
                for i in range(2, 9):
                    comps[i] = int(comps[i]) if comps[i] else -1
                return spwd.struct_spwd(comps)
    raise KeyError


def _getspall(root=None):
    """
    Alternative implementation for getspnam, that use only /etc/shadow
    """
    root = "/" if not root else root
    passwd = os.path.join(root, "etc/shadow")
    with salt.utils.files.fopen(passwd) as fp_:
        for line in fp_:
            comps = line.strip().split(":")
            # Generate a getspall compatible output
            for i in range(2, 9):
                comps[i] = int(comps[i]) if comps[i] else -1
            yield spwd.struct_spwd(comps)



"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>
"""

import pytest

import salt.modules.drac as drac
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {drac: {}}


def test_system_info():
    """
    Tests to return System information
    """
    mock = MagicMock(return_value={"retcode": 0, "stdout": None})
    with patch.dict(drac.__salt__, {"cmd.run_all": mock}):
        mock = MagicMock(return_value="ABC")
        with patch.object(drac, "__parse_drac", mock):
            assert drac.system_info() == "ABC"


def test_network_info():
    """
    Tests to return Network Configuration
    """
    mock = MagicMock(return_value={"retcode": 0, "stdout": None})
    with patch.dict(drac.__salt__, {"cmd.run_all": mock}):
        mock = MagicMock(return_value="ABC")
        with patch.object(drac, "__parse_drac", mock):
            assert drac.network_info() == "ABC"


def test_nameservers():
    """
    tests for configure the nameservers on the DRAC
    """
    assert not drac.nameservers("a", "b", "c")

    mock = MagicMock(return_value=False)
    with patch.object(drac, "__execute_cmd", mock):
        assert not drac.nameservers("a")

    mock = MagicMock(return_value=True)
    with patch.object(drac, "__execute_cmd", mock):
        assert drac.nameservers("a")


def test_syslog():
    """
    Tests for configure syslog remote logging, by default syslog will
    automatically be enabled if a server is specified. However,
    if you want to disable syslog you will need to specify a server
    followed by False
    """
    mock = MagicMock(return_value=True)
    with patch.object(drac, "__execute_cmd", mock):
        assert drac.syslog("server")

    mock = MagicMock(return_value=True)
    with patch.object(drac, "__execute_cmd", mock):
        assert drac.syslog("server", False)


def test_email_alerts():
    """
    Test to Enable/Disable email alerts
    """
    mock = MagicMock(return_value=True)
    with patch.object(drac, "__execute_cmd", mock):
        assert drac.email_alerts(True)

    mock = MagicMock(return_value=True)
    with patch.object(drac, "__execute_cmd", mock):
        assert drac.email_alerts(False)


def test_list_users():
    """
    Test for list all DRAC users
    """
    mock = MagicMock(
        return_value={"retcode": 0, "stdout": "cfgUserAdminUserName=value"}
    )
    with patch.dict(drac.__salt__, {"cmd.run_all": mock}):
        assert drac.list_users() == {"value": {"index": 16}}


def test_delete_user():
    """
    Tests to delete a user
    """
    mock = MagicMock(return_value="ABC")
    with patch.object(drac, "__execute_cmd", mock):
        assert drac.delete_user("username", 1) == "ABC"

    assert not drac.delete_user("username", False)


def test_change_password():
    """
    Tests to change users password
    """
    mock = MagicMock(return_value="ABC")
    with patch.object(drac, "__execute_cmd", mock):
        assert drac.change_password("username", "password", 1) == "ABC"

    assert not drac.change_password("username", "password", False), False


def test_create_user():
    """
    Tests to create user accounts
    """
    assert not drac.create_user(
        "username", "password", "permissions", {"username": None}
    )

    mock = MagicMock(return_value=False)
    with patch.object(drac, "__execute_cmd", mock):
        mock = MagicMock(return_value=None)
        with patch.object(drac, "delete_user", mock):
            assert not drac.create_user(
                "username",
                "password",
                "permissions",
                {"username1": {"index": 1}},
            )

    mock = MagicMock(return_value=True)
    with patch.object(drac, "__execute_cmd", mock):
        mock = MagicMock(return_value=False)
        with patch.object(drac, "set_permissions", mock):
            mock = MagicMock(return_value=None)
            with patch.object(drac, "delete_user", mock):
                assert not drac.create_user(
                    "username",
                    "password",
                    "permissions",
                    {"username1": {"index": 1}},
                )

        mock = MagicMock(return_value=True)
        with patch.object(drac, "set_permissions", mock):
            mock = MagicMock(return_value=False)
            with patch.object(drac, "change_password", mock):
                mock = MagicMock(return_value=None)
                with patch.object(drac, "delete_user", mock):
                    assert not drac.create_user(
                        "username",
                        "password",
                        "permissions",
                        {"username1": {"index": 1}},
                    )

    mock = MagicMock(side_effect=[True, False])
    with patch.object(drac, "__execute_cmd", mock):
        mock = MagicMock(return_value=True)
        with patch.object(drac, "set_permissions", mock):
            mock = MagicMock(return_value=True)
            with patch.object(drac, "change_password", mock):
                mock = MagicMock(return_value=None)
                with patch.object(drac, "delete_user", mock):
                    assert not drac.create_user(
                        "username",
                        "password",
                        "permissions",
                        {"username1": {"index": 1}},
                    )

    mock = MagicMock(side_effect=[True, True])
    with patch.object(drac, "__execute_cmd", mock):
        mock = MagicMock(return_value=True)
        with patch.object(drac, "set_permissions", mock):
            mock = MagicMock(return_value=True)
            with patch.object(drac, "change_password", mock):
                mock = MagicMock(return_value=None)
                with patch.object(drac, "delete_user", mock):
                    assert drac.create_user(
                        "username",
                        "password",
                        "permissions",
                        {"username1": {"index": 1}},
                    )


def test_set_permissions():
    """
    Test to configure users permissions
    """
    mock = MagicMock(return_value=True)
    with patch.object(drac, "__execute_cmd", mock):
        assert drac.set_permissions("username", "A,B,C", 1)


def test_set_snmp():
    """
    Test to configure SNMP community string
    """
    mock = MagicMock(return_value=True)
    with patch.object(drac, "__execute_cmd", mock):
        assert drac.set_snmp("username")


def test_set_network():
    """
    Test to configure Network
    """
    mock = MagicMock(return_value=True)
    with patch.object(drac, "__execute_cmd", mock):
        assert drac.set_network("ip", "netmask", "gateway")


def test_server_reboot():
    """
    Tests for issues a power-cycle operation on the managed server.
    This action is similar to pressing the power button on the system's
    front panel to power down and then power up the system.
    """
    mock = MagicMock(return_value=True)
    with patch.object(drac, "__execute_cmd", mock):
        assert drac.server_reboot()


def test_server_poweroff():
    """
    Tests for powers down the managed server.
    """
    mock = MagicMock(return_value=True)
    with patch.object(drac, "__execute_cmd", mock):
        assert drac.server_poweroff()


def test_server_poweron():
    """
    Tests for powers up the managed server.
    """
    mock = MagicMock(return_value=True)
    with patch.object(drac, "__execute_cmd", mock):
        assert drac.server_poweron()


def test_server_hardreset():
    """
    Tests for performs a reset (reboot) operation on the managed server.
    """
    mock = MagicMock(return_value=True)
    with patch.object(drac, "__execute_cmd", mock):
        assert drac.server_hardreset()


def test_server_pxe():
    """
    Tests to configure server to PXE perform a one off PXE boot
    """
    mock = MagicMock(return_value=True)
    with patch.object(drac, "__execute_cmd", mock):
        assert drac.server_pxe()

    mock = MagicMock(side_effect=[True, False])
    with patch.object(drac, "__execute_cmd", mock):
        assert not drac.server_pxe()

    mock = MagicMock(return_value=False)
    with patch.object(drac, "__execute_cmd", mock):
        assert not drac.server_pxe()



import urllib.parse

import pytest

import salt.utils.json
from salt.ext.tornado.httpclient import HTTPError


async def test_get_root_noauth(http_client):
    """
    GET requests to the root URL should not require auth
    """
    response = await http_client.fetch("/")
    assert response.code == 200


async def test_post_root_auth(http_client):
    """
    POST requests to the root URL redirect to login
    """
    with pytest.raises(HTTPError) as exc:
        await http_client.fetch("/", method="POST", body=salt.utils.json.dumps({}))
    assert exc.value.code == 401


async def test_login_noauth(http_client):
    """
    GET requests to the login URL should not require auth
    """
    response = await http_client.fetch("/login")
    assert response.code == 200


async def test_webhook_auth(http_client):
    """
    Requests to the webhook URL require auth by default
    """
    with pytest.raises(HTTPError) as exc:
        await http_client.fetch("/hook", method="POST", body=salt.utils.json.dumps({}))
    assert exc.value.code == 401


@pytest.mark.destructive_test
@pytest.mark.skip_if_not_root
async def test_good_login(http_client, auth_creds, content_type_map, client_config):
    """
    Test logging in
    """
    response = await http_client.fetch(
        "/login",
        method="POST",
        body=urllib.parse.urlencode(auth_creds),
        headers={"Content-Type": content_type_map["form"]},
    )
    assert response.code == 200
    cookies = response.headers["Set-Cookie"]
    response_obj = salt.utils.json.loads(response.body)["return"][0]
    token = response_obj["token"]
    assert f"session_id={token}" in cookies
    perms = response_obj["perms"]
    perms_config = client_config["external_auth"]["auto"][auth_creds["username"]]
    assert set(perms) == set(perms_config)
    assert "token" in response_obj  # TODO: verify that its valid?
    assert response_obj["user"] == auth_creds["username"]
    assert response_obj["eauth"] == auth_creds["eauth"]


async def test_bad_login(http_client, content_type_map):
    """
    Test logging in
    """
    with pytest.raises(HTTPError) as exc:
        body = urllib.parse.urlencode({"totally": "invalid_creds"})
        await http_client.fetch(
            "/login",
            method="POST",
            body=body,
            headers={"Content-Type": content_type_map["form"]},
        )
    assert exc.value.code == 401


@pytest.mark.destructive_test
@pytest.mark.skip_if_not_root
async def test_logout(http_client, auth_creds, content_type_map):
    response = await http_client.fetch(
        "/login",
        method="POST",
        body=urllib.parse.urlencode(auth_creds),
        headers={"Content-Type": content_type_map["form"]},
    )
    assert response.code == 200
    token = response.headers["X-Auth-Token"]

    body = urllib.parse.urlencode({})
    response = await http_client.fetch(
        "/logout",
        method="POST",
        body=body,
        headers={"content-type": content_type_map["form"], "X-Auth-Token": token},
    )
    assert response.code == 200



"""
A Runner module interface on top of the salt-ssh Python API.

This allows for programmatic use from salt-api, the Reactor, Orchestrate, etc.
"""

import salt.client.ssh.client


def cmd(tgt, fun, arg=(), timeout=None, tgt_type="glob", kwarg=None):
    """
    .. versionadded:: 2015.5.0
    .. versionchanged:: 2017.7.0
        The ``expr_form`` argument has been renamed to ``tgt_type``, earlier
        releases must use ``expr_form``.

    Execute a single command via the salt-ssh subsystem and return all
    routines at once

    A wrapper around the :py:meth:`SSHClient.cmd
    <salt.client.ssh.client.SSHClient.cmd>` method.
    """
    client = salt.client.ssh.client.SSHClient(mopts=__opts__)
    return client.cmd(tgt, fun, arg, timeout, tgt_type, kwarg)



"""
Manage and query Bower packages
===============================

This module manages the installed packages using Bower.
Note that npm, git and bower must be installed for this module to be
available.

"""

import logging
import shlex

import salt.utils.json
import salt.utils.path
from salt.exceptions import CommandExecutionError
from salt.utils.versions import Version

log = logging.getLogger(__name__)

# Function alias to make sure not to shadow built-in's
__func_alias__ = {"list_": "list"}


def __virtual__():
    """
    Only work when Bower is installed
    """
    if salt.utils.path.which("bower") is None:
        return (False, "The bower module could not be loaded: bower command not found")
    return True


def _check_valid_version():
    """
    Check the version of Bower to ensure this module will work. Currently
    bower must be at least version 1.3.
    """
    # pylint: disable=no-member
    bower_version = Version(
        __salt__["cmd.run"]("{} --version".format(salt.utils.path.which("bower")))
    )
    valid_version = Version("1.3")
    # pylint: enable=no-member
    if bower_version < valid_version:
        raise CommandExecutionError(
            "'bower' is not recent enough({} < {}). Please Upgrade.".format(
                bower_version, valid_version
            )
        )


def _construct_bower_command(bower_command):
    """
    Create bower command line string
    """
    if not bower_command:
        raise CommandExecutionError("bower_command, e.g. install, must be specified")

    cmd = [salt.utils.path.which("bower")] + shlex.split(bower_command)
    cmd.extend(
        [
            "--config.analytics",
            "false",
            "--config.interactive",
            "false",
            "--allow-root",
            "--json",
        ]
    )
    return cmd


def install(pkg, dir, pkgs=None, runas=None, env=None):
    """
    Install a Bower package.

    If no package is specified, the dependencies (from bower.json) of the
    package in the given directory will be installed.

    pkg
        A package name in any format accepted by Bower, including a version
        identifier

    dir
        The target directory in which to install the package

    pkgs
        A list of package names in the same format as the ``pkg`` parameter

    runas
        The user to run Bower with

    env
        Environment variables to set when invoking Bower. Uses the same ``env``
        format as the :py:func:`cmd.run <salt.modules.cmdmod.run>` execution
        function.

    CLI Example:

    .. code-block:: bash

        salt '*' bower.install underscore /path/to/project

        salt '*' bower.install jquery#2.0 /path/to/project

    """
    _check_valid_version()

    cmd = _construct_bower_command("install")

    if pkg:
        cmd.append(pkg)
    elif pkgs:
        cmd.extend(pkgs)

    result = __salt__["cmd.run_all"](
        cmd, cwd=dir, runas=runas, env=env, python_shell=False
    )

    if result["retcode"] != 0:
        raise CommandExecutionError(result["stderr"])

    # If package is already installed, Bower will emit empty dict to STDOUT
    stdout = salt.utils.json.loads(result["stdout"])
    return stdout != {}


def uninstall(pkg, dir, runas=None, env=None):
    """
    Uninstall a Bower package.

    pkg
        A package name in any format accepted by Bower

    dir
        The target directory from which to uninstall the package

    runas
        The user to run Bower with

    env
        Environment variables to set when invoking Bower. Uses the same ``env``
        format as the :py:func:`cmd.run <salt.modules.cmdmod.run>` execution
        function.

    CLI Example:

    .. code-block:: bash

        salt '*' bower.uninstall underscore /path/to/project

    """
    _check_valid_version()

    cmd = _construct_bower_command("uninstall")
    cmd.append(pkg)

    result = __salt__["cmd.run_all"](
        cmd, cwd=dir, runas=runas, env=env, python_shell=False
    )

    if result["retcode"] != 0:
        raise CommandExecutionError(result["stderr"])

    # If package is not installed, Bower will emit empty dict to STDOUT
    stdout = salt.utils.json.loads(result["stdout"])
    return stdout != {}


def list_(dir, runas=None, env=None):
    """
    List installed Bower packages.

    dir
        The directory whose packages will be listed

    runas
        The user to run Bower with

    env
        Environment variables to set when invoking Bower. Uses the same ``env``
        format as the :py:func:`cmd.run <salt.modules.cmdmod.run>` execution
        function.

    CLI Example:

    .. code-block:: bash

        salt '*' bower.list /path/to/project

    """
    _check_valid_version()

    cmd = _construct_bower_command("list")
    cmd.append("--offline")

    result = __salt__["cmd.run_all"](
        cmd, cwd=dir, runas=runas, env=env, python_shell=False
    )

    if result["retcode"] != 0:
        raise CommandExecutionError(result["stderr"])

    return salt.utils.json.loads(result["stdout"])["dependencies"]


def prune(dir, runas=None, env=None):
    """
    .. versionadded:: 2017.7.0

    Remove extraneous local Bower packages, i.e. those not referenced in bower.json

    dir
        The directory whose packages will be pruned

    runas
        The user to run Bower with

    env
        Environment variables to set when invoking Bower. Uses the same ``env``
        format as the :py:func:`cmd.run <salt.modules.cmdmod.run>` execution
        function.

    CLI Example:

    .. code-block:: bash

        salt '*' bower.prune /path/to/project

    """
    _check_valid_version()

    cmd = _construct_bower_command("prune")

    result = __salt__["cmd.run_all"](
        cmd, cwd=dir, runas=runas, env=env, python_shell=False
    )

    if result["retcode"] != 0:
        raise CommandExecutionError(result["stderr"])

    # Bower returns an empty dictionary if nothing was pruned
    return salt.utils.json.loads(result["stdout"])



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.modules.twilio_notify as twilio_notify
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase

HAS_LIBS = False
try:
    import twilio

    # Grab version, ensure elements are ints
    twilio_version = tuple(int(x) for x in twilio.__version_info__)
    if twilio_version > (5,):
        TWILIO_5 = False
    else:
        TWILIO_5 = True
    HAS_LIBS = True
except ImportError:
    pass


class MockTwilioRestException(Exception):
    """
    Mock TwilioRestException class
    """

    def __init__(self):
        self.code = "error code"
        self.msg = "Exception error"
        self.status = "Not send"
        super().__init__(self.msg)


class MockMessages:
    """
    Mock SMS class
    """

    flag = None

    def __init__(self):
        self.sid = "011"
        self.price = "200"
        self.price_unit = "1"
        self.status = "Sent"
        self.num_segments = "2"
        self.body = None
        self.date_sent = "01-01-2015"
        self.date_created = "01-01-2015"
        self.to = None
        self.from_ = None

    def create(self, body, to, from_):
        """
        Mock create method
        """
        msg = MockMessages()
        if self.flag == 1:
            raise MockTwilioRestException()
        msg.body = body
        msg.to = to
        msg.from_ = from_
        return msg


class MockSMS:
    """
    Mock SMS class
    """

    def __init__(self):
        self.messages = MockMessages()


class MockTwilioRestClient:
    """
    Mock TwilioRestClient class
    """

    def __init__(self):
        if TWILIO_5:
            self.sms = MockSMS()
        else:
            self.messages = MockMessages()


@pytest.mark.skipif(not HAS_LIBS, reason="twilio.rest is not available")
class TwilioNotifyTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.twilio_notify
    """

    def setup_loader_modules(self):
        return {
            twilio_notify: {
                "TwilioRestClient": MockTwilioRestClient,
                "TwilioRestException": MockTwilioRestException,
            }
        }

    # 'send_sms' function tests: 1

    def test_send_sms(self):
        """
        Test if it send an sms.
        """
        mock = MagicMock(return_value=MockTwilioRestClient())
        with patch.object(twilio_notify, "_get_twilio", mock):
            self.assertDictEqual(
                twilio_notify.send_sms(
                    "twilio-account", "SALTSTACK", "+18019999999", "+18011111111"
                ),
                {
                    "message": {
                        "status": "Sent",
                        "num_segments": "2",
                        "price": "200",
                        "body": "SALTSTACK",
                        "sid": "011",
                        "date_sent": "01-01-2015",
                        "date_created": "01-01-2015",
                        "price_unit": "1",
                    }
                },
            )

            MockMessages.flag = 1
            self.assertDictEqual(
                twilio_notify.send_sms(
                    "twilio-account", "SALTSTACK", "+18019999999", "+18011111111"
                ),
                {
                    "message": {"sid": None},
                    "_error": {
                        "msg": "Exception error",
                        "status": "Not send",
                        "code": "error code",
                    },
                },
            )



"""
Manage VMware distributed virtual switches (DVSs) and their distributed virtual
portgroups (DVportgroups).

:codeauthor: `Alexandru Bleotu <alexandru.bleotu@morganstaley.com>`

Examples
========

Several settings can be changed for DVSs and DVporgroups. Here are two examples
covering all of the settings. Fewer settings can be used

DVS
---

.. code-block:: python

    'name': 'dvs1',
    'max_mtu': 1000,
    'uplink_names': [
        'dvUplink1',
        'dvUplink2',
        'dvUplink3'
    ],
    'capability': {
        'portgroup_operation_supported': false,
        'operation_supported': true,
        'port_operation_supported': false
    },
    'lacp_api_version': 'multipleLag',
    'contact_email': 'foo@email.com',
    'product_info': {
        'version':
        '6.0.0',
        'vendor':
        'VMware,
        Inc.',
        'name':
        'DVS'
    },
    'network_resource_management_enabled': true,
    'contact_name': 'me@email.com',
    'infrastructure_traffic_resource_pools': [
        {
            'reservation': 0,
            'limit': 1000,
            'share_level': 'high',
            'key': 'management',
            'num_shares': 100
        },
        {
            'reservation': 0,
            'limit': -1,
            'share_level': 'normal',
            'key': 'faultTolerance',
            'num_shares': 50
        },
        {
            'reservation': 0,
            'limit': 32000,
            'share_level': 'normal',
            'key': 'vmotion',
            'num_shares': 50
        },
        {
            'reservation': 10000,
            'limit': -1,
            'share_level': 'normal',
            'key': 'virtualMachine',
            'num_shares': 50
        },
        {
            'reservation': 0,
            'limit': -1,
            'share_level': 'custom',
            'key': 'iSCSI',
            'num_shares': 75
        },
        {
            'reservation': 0,
            'limit': -1,
            'share_level': 'normal',
            'key': 'nfs',
            'num_shares': 50
        },
        {
            'reservation': 0,
            'limit': -1,
            'share_level': 'normal',
            'key': 'hbr',
            'num_shares': 50
        },
        {
            'reservation': 8750,
            'limit': 15000,
            'share_level': 'high',
            'key': 'vsan',
            'num_shares': 100
        },
        {
            'reservation': 0,
            'limit': -1,
            'share_level': 'normal',
            'key': 'vdp',
            'num_shares': 50
        }
    ],
    'link_discovery_protocol': {
        'operation':
        'listen',
        'protocol':
        'cdp'
    },
    'network_resource_control_version': 'version3',
    'description': 'Managed by Salt. Random settings.'

Note: The mandatory attribute is: ``name``.

Portgroup
---------

.. code-block:: python

    'security_policy': {
        'allow_promiscuous': true,
        'mac_changes': false,
        'forged_transmits': true
    },
    'name': 'vmotion-v702',
    'out_shaping': {
        'enabled': true,
        'average_bandwidth': 1500,
        'burst_size': 4096,
        'peak_bandwidth': 1500
    },
    'num_ports': 128,
    'teaming': {
        'port_order': {
            'active': [
                'dvUplink2'
            ],
            'standby': [
                'dvUplink1'
            ]
        },
        'notify_switches': false,
        'reverse_policy': true,
        'rolling_order': false,
        'policy': 'failover_explicit',
        'failure_criteria': {
            'check_error_percent': true,
            'full_duplex': false,
            'check_duplex': false,
            'percentage': 50,
            'check_speed': 'minimum',
            'speed': 20,
            'check_beacon': true
        }
    },
    'type': 'earlyBinding',
    'vlan_id': 100,
    'description': 'Managed by Salt. Random settings.'

Note: The mandatory attributes are: ``name``, ``type``.

Dependencies
============

- pyVmomi Python Module


pyVmomi
-------

PyVmomi can be installed via pip:

.. code-block:: bash

    pip install pyVmomi

.. note::

    Version 6.0 of pyVmomi has some problems with SSL error handling on certain
    versions of Python. If using version 6.0 of pyVmomi, Python 2.7.9,
    or newer must be present. This is due to an upstream dependency
    in pyVmomi 6.0 that is not supported in Python versions 2.7 to 2.7.8. If the
    version of Python is not in the supported range, you will need to install an
    earlier version of pyVmomi. See `Issue #29537`_ for more information.

.. _Issue #29537: https://github.com/saltstack/salt/issues/29537

Based on the note above, to install an earlier version of pyVmomi than the
version currently listed in PyPi, run the following:

.. code-block:: bash

    pip install pyVmomi==5.5.0.2014.1.1

The 5.5.0.2014.1.1 is a known stable version that this original ESXi State
Module was developed against.
"""

import logging
import sys

import salt.exceptions

try:
    from pyVmomi import VmomiSupport

    HAS_PYVMOMI = True
except ImportError:
    HAS_PYVMOMI = False

# Get Logging Started
log = logging.getLogger(__name__)


def __virtual__():
    if not HAS_PYVMOMI:
        return False, "State module did not load: pyVmomi not found"

    # We check the supported vim versions to infer the pyVmomi version
    if (
        "vim25/6.0" in VmomiSupport.versionMap
        and sys.version_info > (2, 7)
        and sys.version_info < (2, 7, 9)
    ):

        return (
            False,
            "State module did not load: Incompatible versions "
            "of Python and pyVmomi present. See Issue #29537.",
        )
    return "dvs"


def mod_init(low):
    """
    Init function
    """
    return True


def _get_datacenter_name():
    """
    Returns the datacenter name configured on the proxy

    Supported proxies: esxcluster, esxdatacenter
    """

    proxy_type = __salt__["vsphere.get_proxy_type"]()
    details = None
    if proxy_type == "esxcluster":
        details = __salt__["esxcluster.get_details"]()
    elif proxy_type == "esxdatacenter":
        details = __salt__["esxdatacenter.get_details"]()
    if not details:
        raise salt.exceptions.CommandExecutionError(
            f"details for proxy type '{proxy_type}' not loaded"
        )
    return details["datacenter"]


def dvs_configured(name, dvs):
    """
    Configures a DVS.

    Creates a new DVS, if it doesn't exist in the provided datacenter or
    reconfigures it if configured differently.

    dvs
        DVS dict representations (see module sysdocs)
    """
    datacenter_name = _get_datacenter_name()
    dvs_name = dvs["name"] if dvs.get("name") else name
    log.info(
        "Running state %s for DVS '%s' in datacenter '%s'",
        name,
        dvs_name,
        datacenter_name,
    )
    changes_required = False
    ret = {"name": name, "changes": {}, "result": None, "comment": None}
    comments = []
    changes = {}
    changes_required = False

    try:
        # TODO dvs validation
        si = __salt__["vsphere.get_service_instance_via_proxy"]()
        dvss = __salt__["vsphere.list_dvss"](dvs_names=[dvs_name], service_instance=si)
        if not dvss:
            changes_required = True
            if __opts__["test"]:
                comments.append(
                    "State {} will create a new DVS '{}' in datacenter '{}'".format(
                        name, dvs_name, datacenter_name
                    )
                )
                log.info(comments[-1])
            else:
                dvs["name"] = dvs_name
                __salt__["vsphere.create_dvs"](
                    dvs_dict=dvs, dvs_name=dvs_name, service_instance=si
                )
                comments.append(
                    "Created a new DVS '{}' in datacenter '{}'".format(
                        dvs_name, datacenter_name
                    )
                )
                log.info(comments[-1])
                changes.update({"dvs": {"new": dvs}})
        else:
            # DVS already exists. Checking various aspects of the config
            props = [
                "description",
                "contact_email",
                "contact_name",
                "lacp_api_version",
                "link_discovery_protocol",
                "max_mtu",
                "network_resource_control_version",
                "network_resource_management_enabled",
            ]
            log.trace(
                "DVS '%s' found in datacenter '%s'. Checking for any updates in %s",
                dvs_name,
                datacenter_name,
                props,
            )
            props_to_original_values = {}
            props_to_updated_values = {}
            current_dvs = dvss[0]
            for prop in props:
                if prop in dvs and dvs[prop] != current_dvs.get(prop):
                    props_to_original_values[prop] = current_dvs.get(prop)
                    props_to_updated_values[prop] = dvs[prop]

            # Simple infrastructure traffic resource control compare doesn't
            # work because num_shares is optional if share_level is not custom
            # We need to do a dedicated compare for this property
            infra_prop = "infrastructure_traffic_resource_pools"
            original_infra_res_pools = []
            updated_infra_res_pools = []
            if infra_prop in dvs:
                if not current_dvs.get(infra_prop):
                    updated_infra_res_pools = dvs[infra_prop]
                else:
                    for idx in range(len(dvs[infra_prop])):
                        if (
                            "num_shares" not in dvs[infra_prop][idx]
                            and current_dvs[infra_prop][idx]["share_level"] != "custom"
                            and "num_shares" in current_dvs[infra_prop][idx]
                        ):

                            del current_dvs[infra_prop][idx]["num_shares"]
                        if dvs[infra_prop][idx] != current_dvs[infra_prop][idx]:

                            original_infra_res_pools.append(
                                current_dvs[infra_prop][idx]
                            )
                            updated_infra_res_pools.append(dict(dvs[infra_prop][idx]))
            if updated_infra_res_pools:
                props_to_original_values["infrastructure_traffic_resource_pools"] = (
                    original_infra_res_pools
                )
                props_to_updated_values["infrastructure_traffic_resource_pools"] = (
                    updated_infra_res_pools
                )
            if props_to_updated_values:
                if __opts__["test"]:
                    changes_string = ""
                    for p in props_to_updated_values:
                        if p == "infrastructure_traffic_resource_pools":
                            changes_string += (
                                "\tinfrastructure_traffic_resource_pools:\n"
                            )
                            for idx in range(len(props_to_updated_values[p])):
                                d = props_to_updated_values[p][idx]
                                s = props_to_original_values[p][idx]
                                changes_string += "\t\t{} from '{}' to '{}'\n".format(
                                    d["key"], s, d
                                )
                        else:
                            changes_string += "\t{} from '{}' to '{}'\n".format(
                                p,
                                props_to_original_values[p],
                                props_to_updated_values[p],
                            )
                    comments.append(
                        "State dvs_configured will update DVS '{}' "
                        "in datacenter '{}':\n{}"
                        "".format(dvs_name, datacenter_name, changes_string)
                    )
                    log.info(comments[-1])
                else:
                    __salt__["vsphere.update_dvs"](
                        dvs_dict=props_to_updated_values,
                        dvs=dvs_name,
                        service_instance=si,
                    )
                    comments.append(
                        "Updated DVS '{}' in datacenter '{}'".format(
                            dvs_name, datacenter_name
                        )
                    )
                    log.info(comments[-1])
                changes.update(
                    {
                        "dvs": {
                            "new": props_to_updated_values,
                            "old": props_to_original_values,
                        }
                    }
                )
        __salt__["vsphere.disconnect"](si)
    except salt.exceptions.CommandExecutionError as exc:
        log.error("Error: %s", exc, exc_info=True)
        if si:
            __salt__["vsphere.disconnect"](si)
        if not __opts__["test"]:
            ret["result"] = False
        ret.update(
            {"comment": str(exc), "result": False if not __opts__["test"] else None}
        )
        return ret
    if not comments:
        # We have no changes
        ret.update(
            {
                "comment": (
                    "DVS '{}' in datacenter '{}' is "
                    "correctly configured. Nothing to be done."
                    "".format(dvs_name, datacenter_name)
                ),
                "result": True,
            }
        )
    else:
        ret.update(
            {
                "comment": "\n".join(comments),
                "changes": changes,
                "result": None if __opts__["test"] else True,
            }
        )
    return ret


def _get_diff_dict(dict1, dict2):
    """
    Returns a dictionary with the diffs between two dictionaries

    It will ignore any key that doesn't exist in dict2
    """
    ret_dict = {}
    for p in dict2.keys():
        if p not in dict1:
            ret_dict.update({p: {"val1": None, "val2": dict2[p]}})
        elif dict1[p] != dict2[p]:
            if isinstance(dict1[p], dict) and isinstance(dict2[p], dict):
                sub_diff_dict = _get_diff_dict(dict1[p], dict2[p])
                if sub_diff_dict:
                    ret_dict.update({p: sub_diff_dict})
            else:
                ret_dict.update({p: {"val1": dict1[p], "val2": dict2[p]}})
    return ret_dict


def _get_val2_dict_from_diff_dict(diff_dict):
    """
    Returns a dictionaries with the values stored in val2 of a diff dict.
    """
    ret_dict = {}
    for p in diff_dict.keys():
        if not isinstance(diff_dict[p], dict):
            raise ValueError(f"Unexpected diff difct '{diff_dict}'")
        if "val2" in diff_dict[p].keys():
            ret_dict.update({p: diff_dict[p]["val2"]})
        else:
            ret_dict.update({p: _get_val2_dict_from_diff_dict(diff_dict[p])})
    return ret_dict


def _get_val1_dict_from_diff_dict(diff_dict):
    """
    Returns a dictionaries with the values stored in val1 of a diff dict.
    """
    ret_dict = {}
    for p in diff_dict.keys():
        if not isinstance(diff_dict[p], dict):
            raise ValueError(f"Unexpected diff difct '{diff_dict}'")
        if "val1" in diff_dict[p].keys():
            ret_dict.update({p: diff_dict[p]["val1"]})
        else:
            ret_dict.update({p: _get_val1_dict_from_diff_dict(diff_dict[p])})
    return ret_dict


def _get_changes_from_diff_dict(diff_dict):
    """
    Returns a list of string message of the differences in a diff dict.

    Each inner message is tabulated one tab deeper
    """
    changes_strings = []
    for p in diff_dict.keys():
        if not isinstance(diff_dict[p], dict):
            raise ValueError(f"Unexpected diff difct '{diff_dict}'")
        if sorted(diff_dict[p].keys()) == ["val1", "val2"]:
            # Some string formatting
            from_str = diff_dict[p]["val1"]
            if isinstance(diff_dict[p]["val1"], str):
                from_str = "'{}'".format(diff_dict[p]["val1"])
            elif isinstance(diff_dict[p]["val1"], list):
                from_str = "'{}'".format(", ".join(diff_dict[p]["val1"]))
            to_str = diff_dict[p]["val2"]
            if isinstance(diff_dict[p]["val2"], str):
                to_str = "'{}'".format(diff_dict[p]["val2"])
            elif isinstance(diff_dict[p]["val2"], list):
                to_str = "'{}'".format(", ".join(diff_dict[p]["val2"]))
            changes_strings.append(f"{p} from {from_str} to {to_str}")
        else:
            sub_changes = _get_changes_from_diff_dict(diff_dict[p])
            if sub_changes:
                changes_strings.append(f"{p}:")
                changes_strings.extend([f"\t{c}" for c in sub_changes])
    return changes_strings


def portgroups_configured(name, dvs, portgroups):
    """
    Configures portgroups on a DVS.

    Creates/updates/removes portgroups in a provided DVS

    dvs
        Name of the DVS

    portgroups
        Portgroup dict representations (see module sysdocs)
    """
    datacenter = _get_datacenter_name()
    log.info("Running state %s on DVS '%s', datacenter '%s'", name, dvs, datacenter)
    changes_required = False
    ret = {"name": name, "changes": {}, "result": None, "comment": None}
    comments = []
    changes = {}
    changes_required = False

    try:
        # TODO portroups validation
        si = __salt__["vsphere.get_service_instance_via_proxy"]()
        current_pgs = __salt__["vsphere.list_dvportgroups"](
            dvs=dvs, service_instance=si
        )
        expected_pg_names = []
        for pg in portgroups:
            pg_name = pg["name"]
            expected_pg_names.append(pg_name)
            del pg["name"]
            log.info("Checking pg '%s'", pg_name)
            filtered_current_pgs = [p for p in current_pgs if p.get("name") == pg_name]
            if not filtered_current_pgs:
                changes_required = True
                if __opts__["test"]:
                    comments.append(
                        "State {} will create a new portgroup "
                        "'{}' in DVS '{}', datacenter "
                        "'{}'".format(name, pg_name, dvs, datacenter)
                    )
                else:
                    __salt__["vsphere.create_dvportgroup"](
                        portgroup_dict=pg,
                        portgroup_name=pg_name,
                        dvs=dvs,
                        service_instance=si,
                    )
                    comments.append(
                        "Created a new portgroup '{}' in DVS "
                        "'{}', datacenter '{}'"
                        "".format(pg_name, dvs, datacenter)
                    )
                log.info(comments[-1])
                changes.update({pg_name: {"new": pg}})
            else:
                # Porgroup already exists. Checking the config
                log.trace(
                    "Portgroup '%s' found in DVS '%s', datacenter '%s'. Checking for any updates.",
                    pg_name,
                    dvs,
                    datacenter,
                )
                current_pg = filtered_current_pgs[0]
                diff_dict = _get_diff_dict(current_pg, pg)

                if diff_dict:
                    changes_required = True
                    if __opts__["test"]:
                        changes_strings = _get_changes_from_diff_dict(diff_dict)
                        log.trace("changes_strings = %s", changes_strings)
                        comments.append(
                            "State {} will update portgroup '{}' in "
                            "DVS '{}', datacenter '{}':\n{}"
                            "".format(
                                name,
                                pg_name,
                                dvs,
                                datacenter,
                                "\n".join([f"\t{c}" for c in changes_strings]),
                            )
                        )
                    else:
                        __salt__["vsphere.update_dvportgroup"](
                            portgroup_dict=pg,
                            portgroup=pg_name,
                            dvs=dvs,
                            service_instance=si,
                        )
                        comments.append(
                            "Updated portgroup '{}' in DVS "
                            "'{}', datacenter '{}'"
                            "".format(pg_name, dvs, datacenter)
                        )
                    log.info(comments[-1])
                    changes.update(
                        {
                            pg_name: {
                                "new": _get_val2_dict_from_diff_dict(diff_dict),
                                "old": _get_val1_dict_from_diff_dict(diff_dict),
                            }
                        }
                    )
        # Add the uplink portgroup to the expected pg names
        uplink_pg = __salt__["vsphere.list_uplink_dvportgroup"](
            dvs=dvs, service_instance=si
        )
        expected_pg_names.append(uplink_pg["name"])
        # Remove any extra portgroups
        for current_pg in current_pgs:
            if current_pg["name"] not in expected_pg_names:
                changes_required = True
                if __opts__["test"]:
                    comments.append(
                        "State {} will remove "
                        "the portgroup '{}' from DVS '{}', "
                        "datacenter '{}'"
                        "".format(name, current_pg["name"], dvs, datacenter)
                    )
                else:
                    __salt__["vsphere.remove_dvportgroup"](
                        portgroup=current_pg["name"], dvs=dvs, service_instance=si
                    )
                    comments.append(
                        "Removed the portgroup '{}' from DVS "
                        "'{}', datacenter '{}'"
                        "".format(current_pg["name"], dvs, datacenter)
                    )
                log.info(comments[-1])
                changes.update({current_pg["name"]: {"old": current_pg}})
        __salt__["vsphere.disconnect"](si)
    except salt.exceptions.CommandExecutionError as exc:
        log.error("Error: %s", exc, exc_info=True)
        if si:
            __salt__["vsphere.disconnect"](si)
        if not __opts__["test"]:
            ret["result"] = False
        ret.update(
            {"comment": exc.strerror, "result": False if not __opts__["test"] else None}
        )
        return ret
    if not changes_required:
        # We have no changes
        ret.update(
            {
                "comment": (
                    "All portgroups in DVS '{}', datacenter "
                    "'{}' exist and are correctly configured. "
                    "Nothing to be done.".format(dvs, datacenter)
                ),
                "result": True,
            }
        )
    else:
        ret.update(
            {
                "comment": "\n".join(comments),
                "changes": changes,
                "result": None if __opts__["test"] else True,
            }
        )
    return ret


def uplink_portgroup_configured(name, dvs, uplink_portgroup):
    """
    Configures the uplink portgroup on a DVS. The state assumes there is only
    one uplink portgroup.

    dvs
        Name of the DVS

    upling_portgroup
        Uplink portgroup dict representations (see module sysdocs)

    """
    datacenter = _get_datacenter_name()
    log.info("Running %s on DVS '%s', datacenter '%s'", name, dvs, datacenter)
    changes_required = False
    ret = {"name": name, "changes": {}, "result": None, "comment": None}
    comments = []
    changes = {}
    changes_required = False

    try:
        # TODO portroups validation
        si = __salt__["vsphere.get_service_instance_via_proxy"]()
        current_uplink_portgroup = __salt__["vsphere.list_uplink_dvportgroup"](
            dvs=dvs, service_instance=si
        )
        log.trace("current_uplink_portgroup = %s", current_uplink_portgroup)
        diff_dict = _get_diff_dict(current_uplink_portgroup, uplink_portgroup)
        if diff_dict:
            changes_required = True
            if __opts__["test"]:
                changes_strings = _get_changes_from_diff_dict(diff_dict)
                log.trace("changes_strings = %s", changes_strings)
                comments.append(
                    "State {} will update the "
                    "uplink portgroup in DVS '{}', datacenter "
                    "'{}':\n{}"
                    "".format(
                        name,
                        dvs,
                        datacenter,
                        "\n".join([f"\t{c}" for c in changes_strings]),
                    )
                )
            else:
                __salt__["vsphere.update_dvportgroup"](
                    portgroup_dict=uplink_portgroup,
                    portgroup=current_uplink_portgroup["name"],
                    dvs=dvs,
                    service_instance=si,
                )
                comments.append(
                    "Updated the uplink portgroup in DVS '{}', datacenter '{}'".format(
                        dvs, datacenter
                    )
                )
            log.info(comments[-1])
            changes.update(
                {
                    "uplink_portgroup": {
                        "new": _get_val2_dict_from_diff_dict(diff_dict),
                        "old": _get_val1_dict_from_diff_dict(diff_dict),
                    }
                }
            )
        __salt__["vsphere.disconnect"](si)
    except salt.exceptions.CommandExecutionError as exc:
        log.error("Error: %s", exc, exc_info=True)
        if si:
            __salt__["vsphere.disconnect"](si)
        if not __opts__["test"]:
            ret["result"] = False
        ret.update(
            {"comment": exc.strerror, "result": False if not __opts__["test"] else None}
        )
        return ret
    if not changes_required:
        # We have no changes
        ret.update(
            {
                "comment": (
                    "Uplink portgroup in DVS '{}', datacenter "
                    "'{}' is correctly configured. "
                    "Nothing to be done.".format(dvs, datacenter)
                ),
                "result": True,
            }
        )
    else:
        ret.update(
            {
                "comment": "\n".join(comments),
                "changes": changes,
                "result": None if __opts__["test"] else True,
            }
        )
    return ret



"""RESTCONF
State module for  Proxy minions

:codeauthor: Jamie (Bear) Murphy <jamiemurphyit@gmail.com>
:maturity:   new
:platform:   any

About
=====
This state module was designed to manage RESTCONF states.
This module relies on the RESTCONF proxy module to interface with the devices.
"""

import difflib
import json
import logging
from collections import OrderedDict

import yaml

log = logging.getLogger(__file__)


def __virtual__():
    if "restconf.set_data" in __salt__:
        return True
    return (False, "RESTCONF module could not be loaded")


def config_manage(
    name, path, method, config, init_path=None, init_method="PATCH", init_config=None
):
    """
    Ensure a specific value exists at a given path

    name:
        (str) The name for this rule

    path:
        (str) The RESTCONF path to set / get config

    method:
        (str) rest method to use eg GET, PUT, POST, PATCH, DELETE

    config:
        (dict) The new value at the given path

    init_path: (optional)
        (str) Alternative path incase the path doesnt exist on first pass

    init_method: (optional)
        (str) Method to use on alternative path when setting config, default: PATCH

    init_config: (optional)
        (dict) The new value at the given init path.
        This is only needed if you need to supply a different style of data to an init path.

    Examples:

    .. code-block:: yaml

        do_configure_restconf_endpoint:
          restconf.config_manage:
            - name: random_name_here
            - path: restconf/data/Cisco-IOS-XE-native:native/interface/GigabitEthernet=1%2F0%2F3
            - config:
                Cisco-IOS-XE-native:GigabitEthernet:
                  description: interfaceDescription
                  name: "1/0/3"

    """
    ret = {"result": False, "comment": ""}

    path = str(path)
    name = str(name)
    method = str(method)
    if path == "":
        ret["comment"] = "CRITICAL: path must not be blank"
        log.critical("path must not be blank")
        return ret
    if name == "":
        ret["comment"] = "CRITICAL: name is required"
        log.critical("name is required")
        return ret
    if method == "":
        ret["comment"] = "CRITICAL: method is required"
        log.critical("method is required")
        return ret
    if not isinstance(config, OrderedDict):
        ret["comment"] = "CRITICAL: config must be an OrderedDict type"
        log.critical(
            "config is required, config must be a salt OrderedDict, not a %s",
            type(config),
        )
        return ret
    ret = {"name": name, "result": False, "changes": {}, "comment": ""}

    # TODO: add template function so that config var does not need to be passed

    path_check = __salt__["restconf.path_check"](path, init_path)
    log.debug("path_check:")
    log.debug(path_check)
    if not path_check["result"]:
        ret["result"] = False
        ret["comment"] = "RESTCONF could not find a working PATH to get initial config"
        return ret

    use_conf = config
    if path_check["path_used"] == "init":
        # We will be creating a new endpoint as we are using the init path so config will be blank
        existing = {}
        if init_config is not None:
            # some init paths need a special config layout
            use_conf = init_config
    request_method = method
    if path_check["path_used"] == "init":
        request_method = init_method
        # since we are using the init method we are basicly doing a net new change
        path_check["request_restponse"] = {}

    proposed_config = json.loads(
        json.dumps(use_conf)
    )  # convert from orderedDict to Dict (which is now ordered by default in python3.8)

    log.debug(
        "existing path config(type: %s):\n%s",
        type(path_check["request_restponse"]),
        path_check["request_restponse"],
    )
    log.debug("proposed_config(type: %s):\n%s", type(proposed_config), proposed_config)

    # TODO: migrate the below == check to RecursiveDictDiffer when issue 59017 is fixed
    if path_check["request_restponse"] == proposed_config:
        ret["result"] = True
        ret["comment"] = "Config is already set"

    elif __opts__["test"] is True:
        ret["result"] = None
        ret["changes"]["changed"] = _compare_changes(
            path_check["request_restponse"], proposed_config
        )
        # ret["changes"]["rest_method"] = request_method
        ret["changes"]["rest_method_path"] = path_check["path_used"]
        ret["comment"] = "Config will be added"

    else:
        resp = __salt__["restconf.set_data"](
            path_check["request_path"], request_method, proposed_config
        )

        # Success
        if resp["status"] in [201, 200, 204]:
            ret["result"] = True
            ret["changes"]["changed"] = _compare_changes(
                path_check["request_restponse"], proposed_config
            )
            # ret["changes"]["rest_method"] = request_method
            ret["changes"]["rest_method_path"] = path_check["path_used"]
            ret["comment"] = "Successfully added config"
        else:
            ret["result"] = False
            if "dict" in resp:
                why = resp["dict"]
            elif "body" in resp:
                why = resp["body"]
            else:
                why = None
            ret["comment"] = (
                "failed to add / modify config. "
                "API Statuscode: {}, API Response: {}, URI: {}".format(
                    resp["status"], why, path_check["request_path"]
                )
            )
            log.debug("post_content: %s", json.dumps(proposed_config))

    return ret


def _compare_changes(old, new, output_style="yaml"):
    # option to switch to a json output

    old_raw = yaml.safe_dump(old, default_flow_style=False).splitlines()
    old = [
        " " + x for x in old_raw
    ]  # adding a space to start of each line to make it readable
    new_raw = yaml.safe_dump(new, default_flow_style=False).splitlines()
    new = [
        " " + x for x in new_raw
    ]  # adding a space to start of each line to make it readable
    if output_style == "json":
        old = json.dumps(old, sort_keys=False, indent=2).splitlines()
        new = json.dumps(new, sort_keys=False, indent=2).splitlines()

    diffout = difflib.unified_diff(old, new, fromfile="before", tofile="after")
    diffclean = "\n".join([x.replace("\n", "") for x in diffout])
    log.debug("resconf_diff:")
    log.debug(diffclean)
    return diffclean



"""
============
Windows DACL
============

This salt utility contains objects and functions for setting permissions to
objects in Windows. You can use the built in functions or access the objects
directly to create your own custom functionality. There are two objects, Flags
and Dacl.

If you need access only to flags, use the Flags object.

.. code-block:: python

    import salt.utils.win_dacl
    flags = salt.utils.win_dacl.Flags()
    flag_full_control = flags.ace_perms['file']['basic']['full_control']

The Dacl object inherits Flags. To use the Dacl object:

.. code-block:: python

    import salt.utils.win_dacl
    dacl = salt.utils.win_dacl.Dacl(obj_type='file')
    dacl.add_ace('Administrators', 'grant', 'full_control')
    dacl.save('C:\\temp')

Object types are used by setting the `obj_type` parameter to a valid Windows
object. Valid object types are as follows:

- file
- service
- printer
- registry
- registry32 (for WOW64)
- share

Each object type has its own set up permissions and 'applies to' properties as
follows. At this time only basic permissions are used for setting. Advanced
permissions are listed for displaying the permissions of an object that don't
match the basic permissions, ie. Special permissions. These should match the
permissions you see when you look at the security for an object.

**Basic Permissions**

    ================  ====  ========  =====  =======  =======
    Permissions       File  Registry  Share  Printer  Service
    ================  ====  ========  =====  =======  =======
    full_control      X     X         X               X
    modify            X
    read_execute      X
    read              X     X         X               X
    write             X     X                         X
    read_write                                        X
    change                            X
    print                                    X
    manage_printer                           X
    manage_documents                         X
    ================  ====  ========  =====  =======  =======

**Advanced Permissions**

    =======================  ====  ========  =======  =======
    Permissions              File  Registry  Printer  Service
    =======================  ====  ========  =======  =======
    *** folder permissions
    list_folder              X
    create_files             X
    create_folders           X
    traverse_folder          X
    delete_subfolders_files  X

    *** file permissions
    read_data                X
    write_data               X
    append_data              X
    execute_file             X

    *** common permissions
    read_ea                  X
    write_ea                 X
    read_attributes          X
    write_attributes         X
    delete                   X     X
    read_permissions         X               X        X
    change_permissions       X               X        X
    take_ownership           X               X
    query_value                    X
    set_value                      X
    create_subkey                  X
    enum_subkeys                   X
    notify                         X
    create_link                    X
    read_control                   X
    write_dac                      X
    write_owner                    X
    manage_printer                           X
    print                                    X
    query_config                                      X
    change_config                                     X
    query_status                                      X
    enum_dependents                                   X
    start                                             X
    stop                                              X
    pause_resume                                      X
    interrogate                                       X
    user_defined                                      X
    change_owner                                      X
    =======================  ====  ========  =======  =======

Only the registry and file object types have 'applies to' properties. These
should match what you see when you look at the properties for an object.

    **File types:**

        - this_folder_only: Applies only to this object
        - this_folder_subfolders_files (default): Applies to this object
          and all sub containers and objects
        - this_folder_subfolders: Applies to this object and all sub
          containers, no files
        - this_folder_files: Applies to this object and all file
          objects, no containers
        - subfolders_files: Applies to all containers and objects
          beneath this object
        - subfolders_only: Applies to all containers beneath this object
        - files_only: Applies to all file objects beneath this object

    .. note::

        'applies to' properties can only be modified on directories. Files
        will always be ``this_folder_only``.

    **Registry types:**

        - this_key_only: Applies only to this key
        - this_key_subkeys: Applies to this key and all subkeys
        - subkeys_only: Applies to all subkeys beneath this object

"""

import logging

import salt.utils.platform
import salt.utils.win_functions
from salt.exceptions import CommandExecutionError, SaltInvocationError

HAS_WIN32 = False
try:
    import pywintypes
    import win32api
    import win32con
    import win32security

    HAS_WIN32 = True
except ImportError:
    pass

log = logging.getLogger(__name__)

__virtualname__ = "dacl"


def __virtual__():
    """
    Only load if Win32 Libraries are installed
    """
    if not salt.utils.platform.is_windows():
        return False, "win_dacl: Requires Windows"

    if not HAS_WIN32:
        return False, "win_dacl: Requires pywin32"

    return __virtualname__


def flags(instantiated=True):
    """
    Helper function for instantiating a Flags object

    Args:

        instantiated (bool):
            True to return an instantiated object, False to return the object
            definition. Use False if inherited by another class. Default is
            True.

    Returns:
        object: An instance of the Flags object or its definition
    """
    if not HAS_WIN32:
        return

    class Flags:
        """
        Object containing all the flags for dealing with Windows permissions
        """

        # Flag Dicts
        ace_perms = {
            "file": {
                "basic": {
                    0x1F01FF: "Full control",
                    0x1301BF: "Modify",
                    0x1201BF: "Read & execute with write",
                    0x1200A9: "Read & execute",
                    0x120089: "Read",
                    0x100116: "Write",
                    "full_control": 0x1F01FF,
                    "modify": 0x1301BF,
                    "read_execute": 0x1200A9,
                    "read": 0x120089,
                    "write": 0x100116,
                },
                "advanced": {
                    # Advanced
                    0x0001: "List folder / read data",
                    0x0002: "Create files / write data",
                    0x0004: "Create folders / append data",
                    0x0008: "Read extended attributes",
                    0x0010: "Write extended attributes",
                    0x0020: "Traverse folder / execute file",
                    0x0040: "Delete subfolders and files",
                    0x0080: "Read attributes",
                    0x0100: "Write attributes",
                    0x10000: "Delete",
                    0x20000: "Read permissions",
                    0x40000: "Change permissions",
                    0x80000: "Take ownership",
                    # 0x100000: 'SYNCHRONIZE',  # This is in all of them
                    # Directory permissions
                    "list_folder": 0x0001,
                    "create_files": 0x0002,
                    "create_folders": 0x0004,
                    "traverse_folder": 0x0020,
                    "delete_subfolders_files": 0x0040,
                    # File permissions
                    "read_data": 0x0001,
                    "write_data": 0x0002,
                    "append_data": 0x0004,
                    "execute_file": 0x0020,
                    # Common permissions
                    "read_ea": 0x0008,
                    "write_ea": 0x0010,
                    "read_attributes": 0x0080,
                    "write_attributes": 0x0100,
                    "delete": 0x10000,
                    "read_permissions": 0x20000,
                    "change_permissions": 0x40000,
                    "take_ownership": 0x80000,
                },
            },
            "registry": {
                "basic": {
                    0xF003F: "Full Control",
                    0x20019: "Read",
                    0x20006: "Write",
                    # Generic Values (These sometimes get hit)
                    0x10000000: "Full Control",
                    0x20000000: "Execute",
                    0x40000000: "Write",
                    0xFFFFFFFF80000000: "Read",
                    "full_control": 0xF003F,
                    "read": 0x20019,
                    "write": 0x20006,
                },
                "advanced": {
                    # Advanced
                    0x0001: "Query Value",
                    0x0002: "Set Value",
                    0x0004: "Create Subkey",
                    0x0008: "Enumerate Subkeys",
                    0x0010: "Notify",
                    0x0020: "Create Link",
                    0x10000: "Delete",
                    0x20000: "Read Control",
                    0x40000: "Write DAC",
                    0x80000: "Write Owner",
                    "query_value": 0x0001,
                    "set_value": 0x0002,
                    "create_subkey": 0x0004,
                    "enum_subkeys": 0x0008,
                    "notify": 0x0010,
                    "create_link": 0x0020,
                    "delete": 0x10000,
                    "read_control": 0x20000,
                    "write_dac": 0x40000,
                    "write_owner": 0x80000,
                },
            },
            "share": {
                "basic": {
                    0x1F01FF: "Full control",
                    0x1301BF: "Change",
                    0x1200A9: "Read",
                    "full_control": 0x1F01FF,
                    "change": 0x1301BF,
                    "read": 0x1200A9,
                },
                "advanced": {},  # No 'advanced' for shares, needed for lookup
            },
            "printer": {
                "basic": {
                    0x20008: "Print",
                    0xF000C: "Manage this printer",
                    0xF0030: "Manage documents",
                    "print": 0x20008,
                    "manage_printer": 0xF000C,
                    "manage_documents": 0xF0030,
                },
                "advanced": {
                    # Advanced
                    0x10004: "Manage this printer",
                    0x0008: "Print",
                    0x20000: "Read permissions",
                    0x40000: "Change permissions",
                    0x80000: "Take ownership",
                    "manage_printer": 0x10004,
                    "print": 0x0008,
                    "read_permissions": 0x20000,
                    "change_permissions": 0x40000,
                    "take_ownership": 0x80000,
                },
            },
            "service": {
                "basic": {
                    0xF01FF: "Full Control",
                    0x2008F: "Read & Write",
                    0x2018D: "Read",
                    0x20002: "Write",
                    "full_control": 0xF01FF,
                    "read_write": 0x2008F,
                    "read": 0x2018D,
                    "write": 0x20002,
                },
                "advanced": {
                    0x0001: "Query Config",
                    0x0002: "Change Config",
                    0x0004: "Query Status",
                    0x0008: "Enumerate Dependents",
                    0x0010: "Start",
                    0x0020: "Stop",
                    0x0040: "Pause/Resume",
                    0x0080: "Interrogate",
                    0x0100: "User-Defined Control",
                    # 0x10000: 'Delete',  # Not visible in the GUI
                    0x20000: "Read Permissions",
                    0x40000: "Change Permissions",
                    0x80000: "Change Owner",
                    "query_config": 0x0001,
                    "change_config": 0x0002,
                    "query_status": 0x0004,
                    "enum_dependents": 0x0008,
                    "start": 0x0010,
                    "stop": 0x0020,
                    "pause_resume": 0x0040,
                    "interrogate": 0x0080,
                    "user_defined": 0x0100,
                    "read_permissions": 0x20000,
                    "change_permissions": 0x40000,
                    "change_owner": 0x80000,
                },
            },
        }

        # These denote inheritance
        # 0x0000 : Not inherited, I don't know the enumeration for this
        # 0x0010 : win32security.INHERITED_ACE

        # All the values in the dict below are combinations of the following
        # enumerations or'ed together
        # 0x0001 : win32security.OBJECT_INHERIT_ACE
        # 0x0002 : win32security.CONTAINER_INHERIT_ACE
        # 0x0004 : win32security.NO_PROPAGATE_INHERIT_ACE
        # 0x0008 : win32security.INHERIT_ONLY_ACE

        # These values only apply to directories. Files will always return
        # "This folder only"
        ace_prop = {
            "file": {
                # for report
                0x0000: "This folder only",
                0x0001: "This folder and files",
                0x0002: "This folder and subfolders",
                0x0003: "This folder, subfolders and files",
                0x0009: "Files only",
                0x000A: "Subfolders only",
                0x000B: "Subfolders and files only",
                # for setting
                "this_folder_only": 0x0000,
                "this_folder_files": 0x0001,
                "this_folder_subfolders": 0x0002,
                "this_folder_subfolders_files": 0x0003,
                "files_only": 0x0009,
                "subfolders_only": 0x000A,
                "subfolders_files": 0x000B,
            },
            "registry": {
                0x0000: "This key only",
                0x0002: "This key and subkeys",
                0x000A: "Subkeys only",
                "this_key_only": 0x0000,
                "this_key_subkeys": 0x0002,
                "subkeys_only": 0x000A,
            },
            "registry32": {
                0x0000: "This key only",
                0x0002: "This key and subkeys",
                0x000A: "Subkeys only",
                "this_key_only": 0x0000,
                "this_key_subkeys": 0x0002,
                "subkeys_only": 0x000A,
            },
        }

        ace_type = {
            "grant": win32security.ACCESS_ALLOWED_ACE_TYPE,
            "deny": win32security.ACCESS_DENIED_ACE_TYPE,
            win32security.ACCESS_ALLOWED_ACE_TYPE: "grant",
            win32security.ACCESS_DENIED_ACE_TYPE: "deny",
        }

        element = {
            "dacl": win32security.DACL_SECURITY_INFORMATION,
            "group": win32security.GROUP_SECURITY_INFORMATION,
            "owner": win32security.OWNER_SECURITY_INFORMATION,
        }

        inheritance = {
            "protected": win32security.PROTECTED_DACL_SECURITY_INFORMATION,
            "unprotected": win32security.UNPROTECTED_DACL_SECURITY_INFORMATION,
        }

        obj_type = {
            "file": win32security.SE_FILE_OBJECT,
            "service": win32security.SE_SERVICE,
            "printer": win32security.SE_PRINTER,
            "registry": win32security.SE_REGISTRY_KEY,
            "registry32": win32security.SE_REGISTRY_WOW64_32KEY,
            "share": win32security.SE_LMSHARE,
        }

    return Flags() if instantiated else Flags


def dacl(obj_name=None, obj_type="file"):
    """
    Helper function for instantiating a Dacl class.

    Args:

        obj_name (str):
            The full path to the object. If None, a blank DACL will be created.
            Default is None.

        obj_type (str):
            The type of object. Default is 'File'

    Returns:
        object: An instantiated Dacl object
    """

    if not HAS_WIN32:
        return

    class Dacl(flags(False)):
        """
        DACL Object
        """

        def __init__(self, obj_name=None, obj_type="file"):
            """
            Either load the DACL from the passed object or create an empty DACL.
            If `obj_name` is not passed, an empty DACL is created.

            Args:

                obj_name (str):
                    The full path to the object. If None, a blank DACL will be
                    created

                obj_type (Optional[str]):
                    The type of object.

            Returns:
                obj: A DACL object

            Usage:

            .. code-block:: python

                # Create an Empty DACL
                dacl = Dacl(obj_type=obj_type)

                # Load the DACL of the named object
                dacl = Dacl(obj_name, obj_type)
            """
            # Validate obj_type
            if obj_type.lower() not in self.obj_type:
                raise SaltInvocationError(f'Invalid "obj_type" passed: {obj_type}')

            self.dacl_type = obj_type.lower()

            if obj_name is None:
                self.dacl = win32security.ACL()
            else:
                if "registry" in self.dacl_type:
                    obj_name = self.get_reg_name(obj_name)

                try:
                    sd = win32security.GetNamedSecurityInfo(
                        obj_name, self.obj_type[self.dacl_type], self.element["dacl"]
                    )
                except pywintypes.error as exc:
                    if "The system cannot find" in exc.strerror:
                        msg = f"System cannot find {obj_name}"
                        log.exception(msg)
                        raise CommandExecutionError(msg)
                    raise

                self.dacl = sd.GetSecurityDescriptorDacl()
                if self.dacl is None:
                    self.dacl = win32security.ACL()

        def get_reg_name(self, obj_name):
            """
            Take the obj_name and convert the hive to a valid registry hive.

            Args:

                obj_name (str):
                    The full path to the registry key including the hive, eg:
                    ``HKLM\\SOFTWARE\\salt``. Valid options for the hive are:

                    - HKEY_LOCAL_MACHINE
                    - MACHINE
                    - HKLM
                    - HKEY_USERS
                    - USERS
                    - HKU
                    - HKEY_CURRENT_USER
                    - CURRENT_USER
                    - HKCU
                    - HKEY_CLASSES_ROOT
                    - CLASSES_ROOT
                    - HKCR

            Returns:
                str:
                    The full path to the registry key in the format expected by
                    the Windows API

            Usage:

            .. code-block:: python

                import salt.utils.win_dacl
                dacl = salt.utils.win_dacl.Dacl()
                valid_key = dacl.get_reg_name('HKLM\\SOFTWARE\\salt')

                # Returns: MACHINE\\SOFTWARE\\salt
            """
            # Make sure the hive is correct
            # Should be MACHINE, USERS, CURRENT_USER, or CLASSES_ROOT
            hives = {
                # MACHINE
                "HKEY_LOCAL_MACHINE": "MACHINE",
                "MACHINE": "MACHINE",
                "HKLM": "MACHINE",
                # USERS
                "HKEY_USERS": "USERS",
                "USERS": "USERS",
                "HKU": "USERS",
                # CURRENT_USER
                "HKEY_CURRENT_USER": "CURRENT_USER",
                "CURRENT_USER": "CURRENT_USER",
                "HKCU": "CURRENT_USER",
                # CLASSES ROOT
                "HKEY_CLASSES_ROOT": "CLASSES_ROOT",
                "CLASSES_ROOT": "CLASSES_ROOT",
                "HKCR": "CLASSES_ROOT",
            }
            reg = obj_name.split("\\")
            passed_hive = reg.pop(0)

            try:
                valid_hive = hives[passed_hive.upper()]
            except KeyError:
                log.exception("Invalid Registry Hive: %s", passed_hive)
                raise CommandExecutionError(f"Invalid Registry Hive: {passed_hive}")

            reg.insert(0, valid_hive)

            return r"\\".join(reg)

        def add_ace(self, principal, access_mode, permissions, applies_to):
            """
            Add an ACE to the DACL

            Args:

                principal (str):
                    The sid of the user/group to for the ACE

                access_mode (str):
                    Determines the type of ACE to add. Must be either ``grant``
                    or ``deny``.

                permissions (str, list):
                    The type of permissions to grant/deny the user. Can be one
                    of the basic permissions, or a list of advanced permissions.

                applies_to (str):
                    The objects to which these permissions will apply. Not all
                    these options apply to all object types.

            Returns:
                bool: True if successful, otherwise False

            Usage:

            .. code-block:: python

                dacl = Dacl(obj_type=obj_type)
                dacl.add_ace(sid, access_mode, permission, applies_to)
                dacl.save(obj_name, protected)
            """
            sid = get_sid(principal)

            if applies_to not in self.ace_prop[self.dacl_type]:
                raise SaltInvocationError(
                    f"Invalid 'applies_to' for type {self.dacl_type}"
                )

            if self.dacl is None:
                raise SaltInvocationError("You must load the DACL before adding an ACE")

            # Get the permission flag
            perm_flag = 0
            if isinstance(permissions, str):
                try:
                    perm_flag = self.ace_perms[self.dacl_type]["basic"][permissions]
                except KeyError as exc:
                    msg = f"Invalid permission specified: {permissions}"
                    log.exception(msg)
                    raise CommandExecutionError(msg, exc)
            else:
                for perm in permissions:
                    try:
                        perm_flag |= self.ace_perms[self.dacl_type]["advanced"][perm]
                    except KeyError as exc:
                        msg = f"Invalid permission specified: {perm}"
                        log.exception(msg)
                        raise CommandExecutionError(msg, exc)

            if access_mode.lower() not in ["grant", "deny"]:
                raise SaltInvocationError(f"Invalid Access Mode: {access_mode}")

            # Add ACE to the DACL
            # Grant or Deny
            # There's some strange behavior here with the registry when you give
            # grant permissions to the Administrator account... it adds bit 1 to
            # the propagation flag... so 0x2 becomes 0x3 and results in an
            # Unknown propagation in get_permissions. It displays correctly in
            # the GUI, but after you modify in the GUI they change back to 0x2.
            # Other users work correctly
            try:
                if access_mode.lower() == "grant":
                    self.dacl.AddAccessAllowedAceEx(
                        win32security.ACL_REVISION_DS,
                        # Some types don't support propagation
                        # May need to use 0x0000 instead of None
                        self.ace_prop.get(self.dacl_type, {}).get(applies_to),
                        perm_flag,
                        sid,
                    )
                elif access_mode.lower() == "deny":
                    self.dacl.AddAccessDeniedAceEx(
                        win32security.ACL_REVISION_DS,
                        self.ace_prop.get(self.dacl_type, {}).get(applies_to),
                        perm_flag,
                        sid,
                    )
                else:
                    log.exception("Invalid access mode: %s", access_mode)
                    raise SaltInvocationError(f"Invalid access mode: {access_mode}")
            except Exception as exc:  # pylint: disable=broad-except
                return False, f"Error: {exc}"

            return True

        def order_acl(self):
            """
            Put the ACEs in the ACL in the proper order. This is necessary
            because the add_ace function puts ACEs at the end of the list
            without regard for order. This will cause the following Windows
            Security dialog to appear when viewing the security for the object:

            ``The permissions on Directory are incorrectly ordered, which may
            cause some entries to be ineffective.``

            .. note:: Run this function after adding all your ACEs.

            Proper Orders is as follows:

                1. Implicit Deny
                2. Inherited Deny
                3. Implicit Deny Object
                4. Inherited Deny Object
                5. Implicit Allow
                6. Inherited Allow
                7. Implicit Allow Object
                8. Inherited Allow Object

            Usage:

            .. code-block:: python

                dacl = Dacl(obj_type=obj_type)
                dacl.add_ace(sid, access_mode, applies_to, permission)
                dacl.order_acl()
                dacl.save(obj_name, protected)
            """
            new_dacl = Dacl()
            deny_dacl = Dacl()
            deny_obj_dacl = Dacl()
            allow_dacl = Dacl()
            allow_obj_dacl = Dacl()

            # Load Non-Inherited ACEs first
            for i in range(0, self.dacl.GetAceCount()):
                ace = self.dacl.GetAce(i)
                if ace[0][1] & win32security.INHERITED_ACE == 0:
                    if ace[0][0] == win32security.ACCESS_DENIED_ACE_TYPE:
                        deny_dacl.dacl.AddAccessDeniedAceEx(
                            win32security.ACL_REVISION_DS, ace[0][1], ace[1], ace[2]
                        )
                    elif ace[0][0] == win32security.ACCESS_DENIED_OBJECT_ACE_TYPE:
                        deny_obj_dacl.dacl.AddAccessDeniedAceEx(
                            win32security.ACL_REVISION_DS, ace[0][1], ace[1], ace[2]
                        )
                    elif ace[0][0] == win32security.ACCESS_ALLOWED_ACE_TYPE:
                        allow_dacl.dacl.AddAccessAllowedAceEx(
                            win32security.ACL_REVISION_DS, ace[0][1], ace[1], ace[2]
                        )
                    elif ace[0][0] == win32security.ACCESS_ALLOWED_OBJECT_ACE_TYPE:
                        allow_obj_dacl.dacl.AddAccessAllowedAceEx(
                            win32security.ACL_REVISION_DS, ace[0][1], ace[1], ace[2]
                        )

            # Load Inherited ACEs last
            for i in range(0, self.dacl.GetAceCount()):
                ace = self.dacl.GetAce(i)
                if (
                    ace[0][1] & win32security.INHERITED_ACE
                    == win32security.INHERITED_ACE
                ):
                    ace_prop = ace[0][1] ^ win32security.INHERITED_ACE
                    if ace[0][0] == win32security.ACCESS_DENIED_ACE_TYPE:
                        deny_dacl.dacl.AddAccessDeniedAceEx(
                            win32security.ACL_REVISION_DS, ace_prop, ace[1], ace[2]
                        )
                    elif ace[0][0] == win32security.ACCESS_DENIED_OBJECT_ACE_TYPE:
                        deny_obj_dacl.dacl.AddAccessDeniedAceEx(
                            win32security.ACL_REVISION_DS, ace_prop, ace[1], ace[2]
                        )
                    elif ace[0][0] == win32security.ACCESS_ALLOWED_ACE_TYPE:
                        allow_dacl.dacl.AddAccessAllowedAceEx(
                            win32security.ACL_REVISION_DS, ace_prop, ace[1], ace[2]
                        )
                    elif ace[0][0] == win32security.ACCESS_ALLOWED_OBJECT_ACE_TYPE:
                        allow_obj_dacl.dacl.AddAccessAllowedAceEx(
                            win32security.ACL_REVISION_DS, ace_prop, ace[1], ace[2]
                        )

            # Combine ACEs in the proper order
            # Deny, Deny Object, Allow, Allow Object
            # Deny
            for i in range(0, deny_dacl.dacl.GetAceCount()):
                ace = deny_dacl.dacl.GetAce(i)
                new_dacl.dacl.AddAccessDeniedAceEx(
                    win32security.ACL_REVISION_DS, ace[0][1], ace[1], ace[2]
                )

            # Deny Object
            for i in range(0, deny_obj_dacl.dacl.GetAceCount()):
                ace = deny_obj_dacl.dacl.GetAce(i)
                new_dacl.dacl.AddAccessDeniedAceEx(
                    win32security.ACL_REVISION_DS,
                    ace[0][1] ^ win32security.INHERITED_ACE,
                    ace[1],
                    ace[2],
                )

            # Allow
            for i in range(0, allow_dacl.dacl.GetAceCount()):
                ace = allow_dacl.dacl.GetAce(i)
                new_dacl.dacl.AddAccessAllowedAceEx(
                    win32security.ACL_REVISION_DS, ace[0][1], ace[1], ace[2]
                )

            # Allow Object
            for i in range(0, allow_obj_dacl.dacl.GetAceCount()):
                ace = allow_obj_dacl.dacl.GetAce(i)
                new_dacl.dacl.AddAccessAllowedAceEx(
                    win32security.ACL_REVISION_DS,
                    ace[0][1] ^ win32security.INHERITED_ACE,
                    ace[1],
                    ace[2],
                )

            # Set the new dacl
            self.dacl = new_dacl.dacl

        def get_ace(self, principal):
            """
            Get the ACE for a specific principal.

            Args:

                principal (str):
                    The name of the user or group for which to get the ace. Can
                    also be a SID.

            Returns:
                dict: A dictionary containing the ACEs found for the principal

            Usage:

            .. code-block:: python

                dacl = Dacl(obj_type=obj_type)
                dacl.get_ace()
            """
            principal = get_name(principal)
            aces = self.list_aces()

            # Filter for the principal
            ret = {}
            for inheritance in aces:
                if principal in aces[inheritance]:
                    ret[inheritance] = {principal: aces[inheritance][principal]}

            return ret

        def list_aces(self):
            """
            List all Entries in the dacl.

            Returns:
                dict: A dictionary containing the ACEs for the object

            Usage:

            .. code-block:: python

                dacl = Dacl('C:\\Temp')
                dacl.list_aces()
            """
            ret = {"Inherited": {}, "Not Inherited": {}}

            # loop through each ACE in the DACL
            for i in range(0, self.dacl.GetAceCount()):
                ace = self.dacl.GetAce(i)

                # Get ACE Elements
                user, a_type, a_prop, a_perms, inheritance = self._ace_to_dict(ace)

                if user in ret[inheritance]:
                    ret[inheritance][user][a_type] = {
                        "applies to": a_prop,
                        "permissions": a_perms,
                    }
                else:
                    ret[inheritance][user] = {
                        a_type: {"applies to": a_prop, "permissions": a_perms}
                    }

            return ret

        def _ace_to_dict(self, ace):
            """
            Helper function for creating the ACE return dictionary
            """
            # Get the principal from the sid (object sid)
            sid = win32security.ConvertSidToStringSid(ace[2])
            principal = get_name(sid)

            # Get the ace type
            ace_type = self.ace_type[ace[0][0]]

            # Is the inherited ace flag present
            inherited = ace[0][1] & win32security.INHERITED_ACE == 16

            # If "Only apply these permissions to objects and/or containers
            # within this container" is checked, there is a 0x4 flag set
            container_only = ace[0][1] & win32security.NO_PROPAGATE_INHERIT_ACE == 4

            # Ace Propagation
            ace_prop = "NA"

            # Get the ace propagation properties
            if self.dacl_type in ["file", "registry", "registry32"]:

                ace_prop = ace[0][1]

                # Remove the inherited ace flag and get propagation
                if inherited:
                    ace_prop = ace[0][1] ^ win32security.INHERITED_ACE

                if container_only:
                    ace_prop = ace[0][1] ^ win32security.NO_PROPAGATE_INHERIT_ACE

                # Lookup the propagation
                try:
                    ace_prop = self.ace_prop[self.dacl_type][ace_prop]
                except KeyError:
                    ace_prop = "Unknown propagation"

            # Get the object type
            obj_type = "registry" if self.dacl_type == "registry32" else self.dacl_type

            # Get the ace permissions
            # Check basic permissions first
            ace_perms = self.ace_perms[obj_type]["basic"].get(ace[1], [])

            # If it didn't find basic perms, check advanced permissions
            if not ace_perms:
                ace_perms = []
                for perm in self.ace_perms[obj_type]["advanced"]:
                    # Don't match against the string perms
                    if isinstance(perm, str):
                        continue
                    if ace[1] & perm == perm:
                        ace_perms.append(self.ace_perms[obj_type]["advanced"][perm])
                ace_perms.sort()

            # If still nothing, it must be undefined
            if not ace_perms:
                ace_perms = [f"Undefined Permission: {ace[1]}"]

            return (
                principal,
                ace_type,
                ace_prop,
                ace_perms,
                "Inherited" if inherited else "Not Inherited",
            )

        def rm_ace(self, principal, ace_type="all"):
            """
            Remove a specific ACE from the DACL.

            Args:

                principal (str):
                    The user whose ACE to remove. Can be the user name or a SID.

                ace_type (str):
                    The type of ACE to remove. If not specified, all ACEs will
                    be removed. Default is 'all'. Valid options are:

                    - 'grant'
                    - 'deny'
                    - 'all'

            Returns:
                list: List of removed aces

            Usage:

            .. code-block:: python

                dacl = Dacl(obj_name='C:\\temp', obj_type='file')
                dacl.rm_ace('Users')
                dacl.save(obj_name='C:\\temp')
            """
            sid = get_sid(principal)
            ace_type = ace_type.lower()
            offset = 0
            ret = []

            for i in range(0, self.dacl.GetAceCount()):
                ace = self.dacl.GetAce(i - offset)

                # Is the inherited ace flag present
                inherited = ace[0][1] & win32security.INHERITED_ACE == 16

                if ace[2] == sid and not inherited:
                    if ace_type == "all" or self.ace_type[ace[0][0]] == ace_type:
                        self.dacl.DeleteAce(i - offset)
                        ret.append(self._ace_to_dict(ace))
                        offset += 1

            if not ret:
                ret = [f"ACE not found for {principal}"]

            return ret

        def rm_all_aces(self, ace_type="all"):
            """
            Removes all ACEs from the DACL.

            Args:

                ace_type (str):
                    The type of ACE to remove. If not specified, all ACEs will
                    be removed. Default is 'all'. Valid options are:

                    - 'grant'
                    - 'deny'
                    - 'all'

            Returns:
                list: List of removed aces

            Usage:

            .. code-block:: python

                dacl = Dacl(obj_name='C:\\temp', obj_type='file')
                dacl.rm_all_aces()
                dacl.save(obj_name='C:\\temp')
            """
            offset = 0
            ret = []

            ace_type = ace_type.lower()

            for i in range(0, self.dacl.GetAceCount()):
                ace = self.dacl.GetAce(i - offset)

                # Is the inherited ace flag present
                inherited = ace[0][1] & win32security.INHERITED_ACE == 16

                if not inherited:
                    if ace_type == "all" or self.ace_type[ace[0][0]] == ace_type:
                        self.dacl.DeleteAce(i - offset)
                        ret.append(self._ace_to_dict(ace))
                        offset += 1

            return ret

        def save(self, obj_name, protected=None):
            """
            Save the DACL

            Args:

                obj_name (str):
                    The object for which to set permissions. This can be the
                    path to a file or folder, a registry key, printer, etc. For
                    more information about how to format the name see:

                    https://msdn.microsoft.com/en-us/library/windows/desktop/aa379593(v=vs.85).aspx

                protected (Optional[bool]):
                    True will disable inheritance for the object. False will
                    enable inheritance. None will make no change. Default is
                    ``None``.

            Returns:
                bool: True if successful, Otherwise raises an exception

            Usage:

            .. code-block:: python

                dacl = Dacl(obj_type='file')
                dacl.save('C:\\Temp', True)
            """
            sec_info = self.element["dacl"]

            if protected is not None:
                if protected:
                    sec_info = sec_info | self.inheritance["protected"]
                else:
                    sec_info = sec_info | self.inheritance["unprotected"]

            if self.dacl_type in ["registry", "registry32"]:
                obj_name = self.get_reg_name(obj_name)

            try:
                win32security.SetNamedSecurityInfo(
                    obj_name,
                    self.obj_type[self.dacl_type],
                    sec_info,
                    None,
                    None,
                    self.dacl,
                    None,
                )
            except pywintypes.error as exc:
                raise CommandExecutionError(
                    f"Failed to set permissions: {obj_name}", exc.strerror
                )

            return True

    return Dacl(obj_name, obj_type)


def get_sid(principal):
    """
    Converts a username to a sid, or verifies a sid. Required for working with
    the DACL.

    Args:

        principal(str):
            The principal to lookup the sid. Can be a sid or a username.

    Returns:
        PySID Object: A sid

    Usage:

    .. code-block:: python

        # Get a user's sid
        salt.utils.win_dacl.get_sid('jsnuffy')

        # Verify that the sid is valid
        salt.utils.win_dacl.get_sid('S-1-5-32-544')
    """
    # If None is passed, use the Universal Well-known SID "Null SID"
    if principal is None:
        principal = "NULL SID"

    # Test if the user passed a sid or a name
    try:
        sid = salt.utils.win_functions.get_sid_from_name(principal)
    except CommandExecutionError:
        sid = principal

    # Test if the SID is valid
    try:
        sid = win32security.ConvertStringSidToSid(sid)
    except pywintypes.error:
        log.exception("Invalid user/group or sid: %s", principal)
        raise CommandExecutionError(f"Invalid user/group or sid: {principal}")
    except TypeError:
        raise CommandExecutionError

    return sid


def get_sid_string(principal):
    """
    Converts a PySID object to a string SID.

    Args:

        principal(str):
            The principal to lookup the sid. Must be a PySID object.

    Returns:
        str: A string sid

    Usage:

    .. code-block:: python

        # Get a PySID object
        py_sid = salt.utils.win_dacl.get_sid('jsnuffy')

        # Get the string version of the SID
        salt.utils.win_dacl.get_sid_string(py_sid)
    """
    # If None is passed, use the Universal Well-known SID "Null SID"
    if principal is None:
        principal = "NULL SID"

    try:
        return win32security.ConvertSidToStringSid(principal)
    except TypeError:
        # Not a PySID object
        principal = get_sid(principal)

    try:
        return win32security.ConvertSidToStringSid(principal)
    except pywintypes.error:
        log.exception("Invalid principal %s", principal)
        raise CommandExecutionError(f"Invalid principal {principal}")


def get_name(principal):
    """
    Gets the name from the specified principal.

    Args:

        principal (str):
            Find the Normalized name based on this. Can be a PySID object, a SID
            string, or a username in any capitalization.

            .. note::
                Searching based on the username can be slow on hosts connected
                to large Active Directory domains.

    Returns:
        str: The username that corresponds to the passed principal. If there is
             no corresponding username, the string SID will be returned.
             Capability SIDs will return ``None``.


    Usage:

    .. code-block:: python

        salt.utils.win_dacl.get_name('S-1-5-32-544')
        salt.utils.win_dacl.get_name('adminisTrators')
    """
    # If this is a PySID object, use it
    if isinstance(principal, pywintypes.SIDType):
        sid_obj = principal
    else:
        # If None is passed, use the Universal Well-known SID for "Null SID"
        if principal is None:
            principal = "S-1-0-0"
        # Try Converting String SID to SID Object first as it's least expensive
        try:
            sid_obj = win32security.ConvertStringSidToSid(principal)
        except pywintypes.error:
            # Try Getting the SID Object by Name Lookup last
            # This is expensive, especially on large AD Domains
            try:
                sid_obj = win32security.LookupAccountName(None, principal)[0]
            except pywintypes.error:
                # This is not a PySID object, a SID String, or a valid Account
                # Name. Just pass it and let the LookupAccountSid function try
                # to resolve it
                sid_obj = principal

    # By now we should have a valid PySID object
    str_sid = get_sid_string(sid_obj)

    try:
        name = win32security.LookupAccountSid(None, sid_obj)[0]

        # Let's Check for Virtual Service Accounts
        # Virtual Accounts must be prepended with NT Service in order to resolve
        # properly
        # https://docs.microsoft.com/en-us/previous-versions/technet-magazine/cc138011(v=msdn.10)
        # https://docs.microsoft.com/en-us/previous-versions/windows/it-pro/windows-server-2008-R2-and-2008/dd548356(v=ws.10)
        if str_sid.startswith("S-1-5-80"):
            name = f"NT Service\\{name}"

        return name
    except pywintypes.error as exc:
        # Microsoft introduced the concept of Capability SIDs in Windows 8
        # https://docs.microsoft.com/en-us/windows/security/identity-protection/access-control/security-identifiers#capability-sids
        # https://support.microsoft.com/en-us/help/4502539/some-sids-do-not-resolve-into-friendly-names
        # https://support.microsoft.com/en-us/help/243330/well-known-security-identifiers-in-windows-operating-systems
        # These types of SIDs do not resolve, so we'll just ignore them for this
        # All capability SIDs begin with `S-1-15-3`, so we'll only throw an
        # error when the sid does not begin with `S-1-15-3`
        # 1332: No mapping between account names and security IDs was done
        if exc.winerror == 1332:
            # Capability SID, return None
            if str_sid.startswith("S-1-15-3"):
                log.debug("Name mapping not available for capability SID: %s", str_sid)
                return None

            # User does not exist on the system or is on a disconnected domain
            # Return the SID
            else:
                log.debug(
                    "Could not resolve SID: %s\nThe user has either been removed "
                    "from the system or is a domain user and the system is not "
                    "connected to the domain",
                    str_sid,
                )
                return str_sid

        # Some other unknown error
        else:
            message = f'Error resolving "{principal}: {exc.strerror}"'
            log.exception(message)
            raise CommandExecutionError(message, exc)


def get_owner(obj_name, obj_type="file"):
    r"""
    Gets the owner of the passed object

    Args:

        obj_name (str):
            The path for which to obtain owner information. The format of this
            parameter is different depending on the ``obj_type``

        obj_type (str):
            The type of object to query. This value changes the format of the
            ``obj_name`` parameter as follows:

            - file: indicates a file or directory
                - a relative path, such as ``FileName.txt`` or ``..\FileName``
                - an absolute path, such as ``C:\DirName\FileName.txt``
                - A UNC name, such as ``\\ServerName\ShareName\FileName.txt``
            - service: indicates the name of a Windows service
            - printer: indicates the name of a printer
            - registry: indicates a registry key
                - Uses the following literal strings to denote the hive:
                    - HKEY_LOCAL_MACHINE
                    - MACHINE
                    - HKLM
                    - HKEY_USERS
                    - USERS
                    - HKU
                    - HKEY_CURRENT_USER
                    - CURRENT_USER
                    - HKCU
                    - HKEY_CLASSES_ROOT
                    - CLASSES_ROOT
                    - HKCR
                - Should be in the format of ``HIVE\Path\To\Key``. For example,
                    ``HKLM\SOFTWARE\Windows``
            - registry32: indicates a registry key under WOW64. Formatting is
                the same as it is for ``registry``
            - share: indicates a network share

    Returns:
        str: The owner (group or user)

    Usage:

    .. code-block:: python

        salt.utils.win_dacl.get_owner('c:\\file')
    """
    # Not all filesystems mountable within windows have SecurityDescriptors.
    # For instance, some mounted SAMBA shares, or VirtualBox shared folders. If
    # we can't load a file descriptor for the file, we default to "None"
    # http://support.microsoft.com/kb/243330

    try:
        obj_type_flag = flags().obj_type[obj_type.lower()]
    except KeyError:
        raise SaltInvocationError(f'Invalid "obj_type" passed: {obj_type}')

    if obj_type in ["registry", "registry32"]:
        obj_name = dacl().get_reg_name(obj_name)

    try:
        security_descriptor = win32security.GetNamedSecurityInfo(
            obj_name, obj_type_flag, win32security.OWNER_SECURITY_INFORMATION
        )
        owner_sid = security_descriptor.GetSecurityDescriptorOwner()

    except MemoryError:
        # Generic Memory Error (Windows Server 2003+)
        owner_sid = "S-1-0-0"

    except pywintypes.error as exc:
        # Incorrect function error (Windows Server 2008+)
        if exc.winerror == 1 or exc.winerror == 50:
            owner_sid = "S-1-0-0"
        else:
            log.exception("Failed to get the owner: %s", obj_name)
            raise CommandExecutionError(
                f"Failed to get owner: {obj_name}", exc.strerror
            )

    return get_name(owner_sid)


def get_primary_group(obj_name, obj_type="file"):
    r"""
    Gets the primary group of the passed object

    Args:

        obj_name (str):
            The path for which to obtain primary group information

        obj_type (str):
            The type of object to query. This value changes the format of the
            ``obj_name`` parameter as follows:

            - file: indicates a file or directory
                - a relative path, such as ``FileName.txt`` or ``..\FileName``
                - an absolute path, such as ``C:\DirName\FileName.txt``
                - A UNC name, such as ``\\ServerName\ShareName\FileName.txt``
            - service: indicates the name of a Windows service
            - printer: indicates the name of a printer
            - registry: indicates a registry key
                - Uses the following literal strings to denote the hive:
                    - HKEY_LOCAL_MACHINE
                    - MACHINE
                    - HKLM
                    - HKEY_USERS
                    - USERS
                    - HKU
                    - HKEY_CURRENT_USER
                    - CURRENT_USER
                    - HKCU
                    - HKEY_CLASSES_ROOT
                    - CLASSES_ROOT
                    - HKCR
                - Should be in the format of ``HIVE\Path\To\Key``. For example,
                    ``HKLM\SOFTWARE\Windows``
            - registry32: indicates a registry key under WOW64. Formatting is
                the same as it is for ``registry``
            - share: indicates a network share

    Returns:
        str: The primary group for the object

    Usage:

    .. code-block:: python

        salt.utils.win_dacl.get_primary_group('c:\\file')
    """
    # Not all filesystems mountable within windows have SecurityDescriptors.
    # For instance, some mounted SAMBA shares, or VirtualBox shared folders. If
    # we can't load a file descriptor for the file, we default to "Everyone"
    # http://support.microsoft.com/kb/243330

    # Validate obj_type
    try:
        obj_type_flag = flags().obj_type[obj_type.lower()]
    except KeyError:
        raise SaltInvocationError(f'Invalid "obj_type" passed: {obj_type}')

    if "registry" in obj_type.lower():
        obj_name = dacl().get_reg_name(obj_name)
        log.debug("Name converted to: %s", obj_name)

    try:
        security_descriptor = win32security.GetNamedSecurityInfo(
            obj_name, obj_type_flag, win32security.GROUP_SECURITY_INFORMATION
        )
        primary_group_gid = security_descriptor.GetSecurityDescriptorGroup()

    except MemoryError:
        # Generic Memory Error (Windows Server 2003+)
        primary_group_gid = "S-1-0-0"

    except pywintypes.error as exc:
        # Incorrect function error (Windows Server 2008+)
        if exc.winerror == 1 or exc.winerror == 50:
            primary_group_gid = "S-1-0-0"
        else:
            log.exception("Failed to get the primary group: %s", obj_name)
            raise CommandExecutionError(
                f"Failed to get primary group: {obj_name}", exc.strerror
            )

    return get_name(win32security.ConvertSidToStringSid(primary_group_gid))


def set_owner(obj_name, principal, obj_type="file"):
    """
    Set the owner of an object. This can be a file, folder, registry key,
    printer, service, etc...

    Args:

        obj_name (str):
            The object for which to set owner. This can be the path to a file or
            folder, a registry key, printer, etc. For more information about how
            to format the name see:

            https://msdn.microsoft.com/en-us/library/windows/desktop/aa379593(v=vs.85).aspx

        principal (str):
            The name of the user or group to make owner of the object. Can also
            pass a SID.

        obj_type (Optional[str]):
            The type of object for which to set the owner. Default is ``file``

    Returns:
        bool: True if successful, raises an error otherwise

    Usage:

    .. code-block:: python

        salt.utils.win_dacl.set_owner('C:\\MyDirectory', 'jsnuffy', 'file')
    """
    sid = get_sid(principal)

    obj_flags = flags()

    # Validate obj_type
    if obj_type.lower() not in obj_flags.obj_type:
        raise SaltInvocationError(f'Invalid "obj_type" passed: {obj_type}')

    if "registry" in obj_type.lower():
        obj_name = dacl().get_reg_name(obj_name)

    # To set the owner to something other than the logged in user requires
    # SE_TAKE_OWNERSHIP_NAME and SE_RESTORE_NAME privileges
    # Enable them for the logged in user
    # Setup the privilege set
    new_privs = set()
    luid = win32security.LookupPrivilegeValue("", "SeTakeOwnershipPrivilege")
    new_privs.add((luid, win32con.SE_PRIVILEGE_ENABLED))
    luid = win32security.LookupPrivilegeValue("", "SeRestorePrivilege")
    new_privs.add((luid, win32con.SE_PRIVILEGE_ENABLED))

    # Get the current token
    p_handle = win32api.GetCurrentProcess()
    t_handle = win32security.OpenProcessToken(
        p_handle, win32security.TOKEN_ALL_ACCESS | win32con.TOKEN_ADJUST_PRIVILEGES
    )

    # Enable the privileges
    win32security.AdjustTokenPrivileges(t_handle, 0, new_privs)

    # Set the user
    try:
        win32security.SetNamedSecurityInfo(
            obj_name,
            obj_flags.obj_type[obj_type.lower()],
            obj_flags.element["owner"],
            sid,
            None,
            None,
            None,
        )
    except pywintypes.error as exc:
        log.exception("Failed to make %s the owner: %s", principal, exc)
        raise CommandExecutionError(f"Failed to set owner: {obj_name}", exc.strerror)

    return True


def set_primary_group(obj_name, principal, obj_type="file"):
    """
    Set the primary group of an object. This can be a file, folder, registry
    key, printer, service, etc...

    Args:

        obj_name (str):
            The object for which to set primary group. This can be the path to a
            file or folder, a registry key, printer, etc. For more information
            about how to format the name see:

            https://msdn.microsoft.com/en-us/library/windows/desktop/aa379593(v=vs.85).aspx

        principal (str):
            The name of the group to make primary for the object. Can also pass
            a SID.

        obj_type (Optional[str]):
            The type of object for which to set the primary group.

    Returns:
        bool: True if successful, raises an error otherwise

    Usage:

    .. code-block:: python

        salt.utils.win_dacl.set_primary_group('C:\\MyDirectory', 'Administrators', 'file')
    """
    # Windows has the concept of a group called 'None'. It is the default group
    # for all Objects. If the user passes None, assume 'None'
    if principal is None:
        principal = "None"

    gid = get_sid(principal)

    obj_flags = flags()

    # Validate obj_type
    if obj_type.lower() not in obj_flags.obj_type:
        raise SaltInvocationError(f'Invalid "obj_type" passed: {obj_type}')

    if "registry" in obj_type.lower():
        obj_name = dacl().get_reg_name(obj_name)

    # To set the owner to something other than the logged in user requires
    # SE_TAKE_OWNERSHIP_NAME and SE_RESTORE_NAME privileges
    # Enable them for the logged in user
    # Setup the privilege set
    new_privs = set()
    luid = win32security.LookupPrivilegeValue("", "SeTakeOwnershipPrivilege")
    new_privs.add((luid, win32con.SE_PRIVILEGE_ENABLED))
    luid = win32security.LookupPrivilegeValue("", "SeRestorePrivilege")
    new_privs.add((luid, win32con.SE_PRIVILEGE_ENABLED))

    # Get the current token
    p_handle = win32api.GetCurrentProcess()
    t_handle = win32security.OpenProcessToken(
        p_handle, win32security.TOKEN_ALL_ACCESS | win32con.TOKEN_ADJUST_PRIVILEGES
    )

    # Enable the privileges
    win32security.AdjustTokenPrivileges(t_handle, 0, new_privs)

    # Set the user
    try:
        win32security.SetNamedSecurityInfo(
            obj_name,
            obj_flags.obj_type[obj_type.lower()],
            obj_flags.element["group"],
            None,
            gid,
            None,
            None,
        )
    except pywintypes.error as exc:
        log.exception("Failed to make %s the primary group: %s", principal, exc)
        raise CommandExecutionError(
            f"Failed to set primary group: {obj_name}", exc.strerror
        )

    return True


def set_permissions(
    obj_name,
    principal,
    permissions,
    access_mode="grant",
    applies_to=None,
    obj_type="file",
    reset_perms=False,
    protected=None,
):
    """
    Set the permissions of an object. This can be a file, folder, registry key,
    printer, service, etc...

    Args:

        obj_name (str):
            The object for which to set permissions. This can be the path to a
            file or folder, a registry key, printer, etc. For more information
            about how to format the name see:

            https://msdn.microsoft.com/en-us/library/windows/desktop/aa379593(v=vs.85).aspx

        principal (str):
            The name of the user or group for which to set permissions. Can also
            pass a SID.

        permissions (str, list):
            The type of permissions to grant/deny the user. Can be one of the
            basic permissions, or a list of advanced permissions.

        access_mode (Optional[str]):
            Whether to grant or deny user the access. Valid options are:

            - grant (default): Grants the user access
            - deny: Denies the user access

        applies_to (Optional[str]):
            The objects to which these permissions will apply. Not all these
            options apply to all object types. Defaults to
            'this_folder_subfolders_files'

        obj_type (Optional[str]):
            The type of object for which to set permissions. Default is 'file'

        reset_perms (Optional[bool]):
            True will overwrite the permissions on the specified object. False
            will append the permissions. Default is False

        protected (Optional[bool]):
            True will disable inheritance for the object. False will enable
            inheritance. None will make no change. Default is None.

    Returns:
        bool: True if successful, raises an error otherwise

    Usage:

    .. code-block:: python

        salt.utils.win_dacl.set_permissions(
            'C:\\Temp', 'jsnuffy', 'full_control', 'grant')
    """
    # Set up applies_to defaults used by registry and file types
    if applies_to is None:
        if "registry" in obj_type.lower():
            applies_to = "this_key_subkeys"
        elif obj_type.lower() == "file":
            applies_to = "this_folder_subfolders_files"

    # If reset_perms is true, it will create a blank DACL
    # Otherwise, it will grab the existing DACL and add to it
    if reset_perms:
        obj_dacl = dacl(obj_type=obj_type)
    else:
        obj_dacl = dacl(obj_name, obj_type)
        obj_dacl.rm_ace(principal, access_mode)

    obj_dacl.add_ace(principal, access_mode, permissions, applies_to)

    obj_dacl.order_acl()

    obj_dacl.save(obj_name, protected)

    return True


def rm_permissions(obj_name, principal, ace_type="all", obj_type="file"):
    r"""
    Remove a user's ACE from an object. This can be a file, folder, registry
    key, printer, service, etc...

    Args:

        obj_name (str):
            The object from which to remove the ace. This can be the
            path to a file or folder, a registry key, printer, etc. For more
            information about how to format the name see:

            https://msdn.microsoft.com/en-us/library/windows/desktop/aa379593(v=vs.85).aspx

        principal (str):
            The name of the user or group for which to set permissions. Can also
            pass a SID.

        ace_type (Optional[str]):
            The type of ace to remove. There are two types of ACEs, 'grant' and
            'deny'. 'all' will remove all ACEs for the user. Default is 'all'

        obj_type (Optional[str]):
            The type of object for which to set permissions. Default is 'file'

    Returns:
        bool: True if successful, raises an error otherwise

    Usage:

    .. code-block:: python

        # Remove jsnuffy's grant ACE from C:\Temp
        salt.utils.win_dacl.rm_permissions('C:\\Temp', 'jsnuffy', 'grant')

        # Remove all ACEs for jsnuffy from C:\Temp
        salt.utils.win_dacl.rm_permissions('C:\\Temp', 'jsnuffy')
    """
    obj_dacl = dacl(obj_name, obj_type)

    obj_dacl.rm_ace(principal, ace_type)
    obj_dacl.save(obj_name)

    return True


def get_permissions(obj_name, principal=None, obj_type="file"):
    """
    Get the permissions for the passed object

    Args:

        obj_name (str):
            The name of or path to the object.

        principal (Optional[str]):
            The name of the user or group for which to get permissions. Can also
            pass a SID. If None, all ACEs defined on the object will be
            returned. Default is None

        obj_type (Optional[str]):
            The type of object for which to get permissions.

    Returns:
        dict: A dictionary representing the object permissions

    Usage:

    .. code-block:: python

        salt.utils.win_dacl.get_permissions('C:\\Temp')
    """
    obj_dacl = dacl(obj_name=obj_name, obj_type=obj_type)

    if principal is None:
        return obj_dacl.list_aces()

    return obj_dacl.get_ace(principal)


def has_permission(
    obj_name, principal, permission, access_mode="grant", obj_type="file", exact=True
):
    r"""
    Check if the object has a specific permission

    Args:

        obj_name (str):
            The name of or path to the object.

        principal (str):
            The name of the user or group for which to get permissions. Can also
            pass a SID.

        permission (str):
            The permission to verify. Valid options depend on the obj_type.

        access_mode (Optional[str]):
            The access mode to check. Is the user granted or denied the
            permission. Default is 'grant'. Valid options are:

            - grant
            - deny

        obj_type (Optional[str]):
            The type of object for which to check permissions. Default is 'file'

        exact (Optional[bool]):
            True for an exact match, otherwise check to see if the permission is
            included in the ACE. Default is True

    Returns:
        bool: True if the object has the permission, otherwise False

    Usage:

    .. code-block:: python

        # Does Joe have read permissions to C:\Temp
        salt.utils.win_dacl.has_permission('C:\\Temp', 'joe', 'read', 'grant', exact=False)

        # Does Joe have Full Control of C:\Temp
        salt.utils.win_dacl.has_permission('C:\\Temp', 'joe', 'full_control', 'grant')
    """
    # Validate access_mode
    if access_mode.lower() not in ["grant", "deny"]:
        raise SaltInvocationError(f'Invalid "access_mode" passed: {access_mode}')
    access_mode = access_mode.lower()

    # Get the DACL
    obj_dacl = dacl(obj_name, obj_type)

    obj_type = obj_type.lower()

    # Get a PySID object
    sid = get_sid(principal)

    # Get the passed permission flag, check basic first
    chk_flag = obj_dacl.ace_perms[obj_type]["basic"].get(
        permission.lower(),
        obj_dacl.ace_perms[obj_type]["advanced"].get(permission.lower(), False),
    )
    if not chk_flag:
        raise SaltInvocationError(f'Invalid "permission" passed: {permission}')

    # Check each ace for sid and type
    cur_flag = None
    for i in range(0, obj_dacl.dacl.GetAceCount()):
        ace = obj_dacl.dacl.GetAce(i)
        if ace[2] == sid and obj_dacl.ace_type[ace[0][0]] == access_mode:
            cur_flag = ace[1]

    # If the ace is empty, return false
    if not cur_flag:
        return False

    # Check if the ACE contains the exact flag
    if exact:
        return cur_flag == chk_flag

    # Check if the ACE contains the permission
    return cur_flag & chk_flag == chk_flag


def has_permissions(
    obj_name, principal, permissions, access_mode="grant", obj_type="file", exact=True
):
    r"""
    Check if the object has the passed permissions. Can be all them or the exact
    permissions passed and nothing more.

    Args:

        obj_name (str):
            The name of or path to the object.

        principal (str):
            The name of the user or group for which to get permissions. Can also
            pass a SID.

        permissions (list):
            The list of permissions to verify

        access_mode (Optional[str]):
            The access mode to check. Is the user granted or denied the
            permission. Default is 'grant'. Valid options are:

            - grant
            - deny

        obj_type (Optional[str]):
            The type of object for which to check permissions. Default is 'file'

        exact (Optional[bool]):
            ``True`` checks if the permissions are exactly those passed in
            permissions. ``False`` checks to see if the permissions are included
            in the ACE. Default is ``True``

    Returns:
        bool: True if the object has the permission, otherwise False

    Usage:

    .. code-block:: python

        # Does Joe have read and write permissions to C:\Temp
        salt.utils.win_dacl.has_permission('C:\\Temp', 'joe', ['read', 'write'], 'grant', exact=False)

        # Does Joe have Full Control of C:\Temp
        salt.utils.win_dacl.has_permissions('C:\\Temp', 'joe', 'full_control', 'grant')
    """
    # If this is a single permission, use has_permission function
    if isinstance(permissions, str):
        return has_permission(
            obj_name=obj_name,
            obj_type=obj_type,
            permission=permissions,
            access_mode=access_mode,
            principal=principal,
            exact=exact,
        )

    # Validate access_mode
    if access_mode.lower() not in ["grant", "deny"]:
        raise SaltInvocationError(f'Invalid "access_mode" passed: {access_mode}')
    access_mode = access_mode.lower()

    # Get the DACL
    obj_dacl = dacl(obj_name, obj_type)

    obj_type = obj_type.lower()

    # Get a PySID object
    sid = get_sid(principal)

    # Get the passed permission flag, check basic first
    chk_flag = 0x0
    for permission in permissions:
        chk_flag |= obj_dacl.ace_perms[obj_type]["basic"].get(
            permission.lower(),
            obj_dacl.ace_perms[obj_type]["advanced"].get(permission.lower(), False),
        )
        if not chk_flag:
            raise SaltInvocationError(f'Invalid "permission" passed: {permission}')

    # Check each ace for sid and type
    cur_flag = None
    for i in range(0, obj_dacl.dacl.GetAceCount()):
        ace = obj_dacl.dacl.GetAce(i)
        if ace[2] == sid and obj_dacl.ace_type[ace[0][0]] == access_mode:
            cur_flag = ace[1]

    # If the ace is empty, return false
    if not cur_flag:
        return False

    # Check if the ACE contains the exact flag
    if exact:
        return cur_flag == chk_flag

    # Check if the ACE contains the permission
    return cur_flag & chk_flag == chk_flag


def set_inheritance(obj_name, enabled, obj_type="file", clear=False):
    """
    Enable or disable an objects inheritance.

    Args:

        obj_name (str):
            The name of the object

        enabled (bool):
            True to enable inheritance, False to disable

        obj_type (Optional[str]):
            The type of object. Only three objects allow inheritance. Valid
            objects are:

            - file (default): This is a file or directory
            - registry
            - registry32 (for WOW64)

        clear (Optional[bool]):
            True to clear existing ACEs, False to keep existing ACEs.
            Default is False

    Returns:
        bool: True if successful, otherwise an Error

    Usage:

    .. code-block:: python

        salt.utils.win_dacl.set_inheritance('C:\\Temp', False)
    """
    if obj_type not in ["file", "registry", "registry32"]:
        raise SaltInvocationError(
            f"obj_type called with incorrect parameter: {obj_name}"
        )

    if clear:
        obj_dacl = dacl(obj_type=obj_type)
    else:
        obj_dacl = dacl(obj_name, obj_type)

    return obj_dacl.save(obj_name, not enabled)


def get_inheritance(obj_name, obj_type="file"):
    """
    Get an object's inheritance.

    Args:

        obj_name (str):
            The name of the object

        obj_type (Optional[str]):
            The type of object. Only three object types allow inheritance. Valid
            objects are:

            - file (default): This is a file or directory
            - registry
            - registry32 (for WOW64)

            The following should return False as there is no inheritance:

            - service
            - printer
            - share

    Returns:
        bool: True if enabled, otherwise False

    Usage:

    .. code-block:: python

        salt.utils.win_dacl.get_inheritance('HKLM\\SOFTWARE\\salt', 'registry')
    """
    obj_dacl = dacl(obj_name=obj_name, obj_type=obj_type)
    inherited = win32security.INHERITED_ACE

    for i in range(0, obj_dacl.dacl.GetAceCount()):
        ace = obj_dacl.dacl.GetAce(i)
        if ace[0][1] & inherited == inherited:
            return True

    return False


def copy_security(
    source,
    target,
    obj_type="file",
    copy_owner=True,
    copy_group=True,
    copy_dacl=True,
    copy_sacl=True,
):
    r"""
    Copy the security descriptor of the Source to the Target. You can specify a
    specific portion of the security descriptor to copy using one of the
    `copy_*` parameters.

    .. note::
        At least one `copy_*` parameter must be ``True``

    .. note::
        The user account running this command must have the following
        privileges:

        - SeTakeOwnershipPrivilege
        - SeRestorePrivilege
        - SeSecurityPrivilege

    Args:

        source (str):
            The full path to the source. This is where the security info will be
            copied from

        target (str):
            The full path to the target. This is where the security info will be
            applied

        obj_type (str): file
            The type of object to query. This value changes the format of the
            ``obj_name`` parameter as follows:
            - file: indicates a file or directory
                - a relative path, such as ``FileName.txt`` or ``..\FileName``
                - an absolute path, such as ``C:\DirName\FileName.txt``
                - A UNC name, such as ``\\ServerName\ShareName\FileName.txt``
            - service: indicates the name of a Windows service
            - printer: indicates the name of a printer
            - registry: indicates a registry key
                - Uses the following literal strings to denote the hive:
                    - HKEY_LOCAL_MACHINE
                    - MACHINE
                    - HKLM
                    - HKEY_USERS
                    - USERS
                    - HKU
                    - HKEY_CURRENT_USER
                    - CURRENT_USER
                    - HKCU
                    - HKEY_CLASSES_ROOT
                    - CLASSES_ROOT
                    - HKCR
                - Should be in the format of ``HIVE\Path\To\Key``. For example,
                    ``HKLM\SOFTWARE\Windows``
            - registry32: indicates a registry key under WOW64. Formatting is
                the same as it is for ``registry``
            - share: indicates a network share

        copy_owner (bool): True
            ``True`` copies owner information. Default is ``True``

        copy_group (bool): True
            ``True`` copies group information. Default is ``True``

        copy_dacl (bool): True
            ``True`` copies the DACL. Default is ``True``

        copy_sacl (bool): True
            ``True`` copies the SACL. Default is ``True``

    Returns:
        bool: ``True`` if successful

    Raises:
        SaltInvocationError: When parameters are invalid
        CommandExecutionError: On failure to set security

    Usage:

    .. code-block:: python

        salt.utils.win_dacl.copy_security(
            source='C:\\temp\\source_file.txt',
            target='C:\\temp\\target_file.txt',
            obj_type='file')

        salt.utils.win_dacl.copy_security(
            source='HKLM\\SOFTWARE\\salt\\test_source',
            target='HKLM\\SOFTWARE\\salt\\test_target',
            obj_type='registry',
            copy_owner=False)
    """
    obj_dacl = dacl(obj_type=obj_type)
    if "registry" in obj_type.lower():
        source = obj_dacl.get_reg_name(source)
        log.info("Source converted to: %s", source)
        target = obj_dacl.get_reg_name(target)
        log.info("Target converted to: %s", target)

    # Set flags
    try:
        obj_type_flag = flags().obj_type[obj_type.lower()]
    except KeyError:
        raise SaltInvocationError(f'Invalid "obj_type" passed: {obj_type}')

    security_flags = 0
    if copy_owner:
        security_flags |= win32security.OWNER_SECURITY_INFORMATION
    if copy_group:
        security_flags |= win32security.GROUP_SECURITY_INFORMATION
    if copy_dacl:
        security_flags |= win32security.DACL_SECURITY_INFORMATION
    if copy_sacl:
        security_flags |= win32security.SACL_SECURITY_INFORMATION

    if not security_flags:
        raise SaltInvocationError(
            "One of copy_owner, copy_group, copy_dacl, or copy_sacl must be True"
        )

    # To set the owner to something other than the logged in user requires
    # SE_TAKE_OWNERSHIP_NAME and SE_RESTORE_NAME privileges
    # Enable them for the logged in user
    # Setup the privilege set
    new_privs = set()
    luid = win32security.LookupPrivilegeValue("", "SeTakeOwnershipPrivilege")
    new_privs.add((luid, win32con.SE_PRIVILEGE_ENABLED))
    luid = win32security.LookupPrivilegeValue("", "SeRestorePrivilege")
    new_privs.add((luid, win32con.SE_PRIVILEGE_ENABLED))
    luid = win32security.LookupPrivilegeValue("", "SeSecurityPrivilege")
    new_privs.add((luid, win32con.SE_PRIVILEGE_ENABLED))

    # Get the current token
    p_handle = win32api.GetCurrentProcess()
    t_handle = win32security.OpenProcessToken(
        p_handle, win32security.TOKEN_ALL_ACCESS | win32con.TOKEN_ADJUST_PRIVILEGES
    )

    # Enable the privileges
    win32security.AdjustTokenPrivileges(t_handle, 0, new_privs)

    # Load object Security Info from the Source
    sec = win32security.GetNamedSecurityInfo(source, obj_type_flag, security_flags)

    # The following return None if the corresponding flag is not set
    sd_sid = sec.GetSecurityDescriptorOwner()
    sd_gid = sec.GetSecurityDescriptorGroup()
    sd_dacl = sec.GetSecurityDescriptorDacl()
    sd_sacl = sec.GetSecurityDescriptorSacl()

    # Set Security info on the target
    try:
        win32security.SetNamedSecurityInfo(
            target, obj_type_flag, security_flags, sd_sid, sd_gid, sd_dacl, sd_sacl
        )
    except pywintypes.error as exc:
        raise CommandExecutionError(f"Failed to set security info: {exc.strerror}")

    return True


def _check_perms(obj_name, obj_type, new_perms, access_mode, ret, test_mode=False):
    """
    Helper function used by ``check_perms`` for checking and setting Grant and
    Deny permissions.

    Args:

        obj_name (str):
            The name or full path to the object

        obj_type (Optional[str]):
            The type of object for which to check permissions. Default is 'file'

        new_perms (dict):
            A dictionary containing the user/group and the basic permissions to
            check/grant, ie: ``{'user': {'perms': 'basic_permission'}}``.

        access_mode (str):
            The access mode to set. Either ``grant`` or ``deny``

        ret (dict):
            A dictionary to append changes to and return. If not passed, will
            create a new dictionary to return.

        test_mode (bool):
            ``True`` will only return the changes that would be made. ``False``
            will make the changes as well as return the changes that would be
            made.

    Returns:
        dict: A dictionary of return data as expected by the state system
    """
    access_mode = access_mode.lower()
    perms_label = f"{access_mode}_perms"
    cur_perms = get_permissions(obj_name=obj_name, obj_type=obj_type)
    changes = {}
    for user in new_perms:
        # Check that user exists:
        user_name = get_name(principal=user)
        # username will be the SID if there is no corresponding username
        if user_name == get_sid_string(principal=user):
            ret["comment"].append(
                "{} Perms: Could not find a corresponding username for: {}".format(
                    access_mode.capitalize(), user
                )
            )
            continue
        if user_name is None:
            ret["comment"].append(
                "{} Perms: Skipping Capability SID: {}".format(
                    access_mode.capitalize(), user
                )
            )
            continue

        # Get the proper applies_to text
        if user_name not in cur_perms["Not Inherited"]:
            changes.setdefault(user, {})
            changes[user]["permissions"] = new_perms[user]["perms"]
            if "applies_to" in new_perms[user]:
                changes[user]["applies_to"] = new_perms[user]["applies_to"]
        else:
            # Check existing permissions
            if not has_permissions(
                obj_name=obj_name,
                principal=user_name,
                permissions=new_perms[user]["perms"],
                access_mode=access_mode,
                obj_type=obj_type,
                exact=True,
            ):
                changes.setdefault(user, {})
                changes[user]["permissions"] = new_perms[user]["perms"]

                # Check existing propagation
                if "applies_to" in new_perms[user]:
                    applies_to = new_perms[user]["applies_to"]
                    at_flag = flags().ace_prop[obj_type][applies_to]
                    applies_to_text = flags().ace_prop[obj_type][at_flag]

                    # Is there a deny/grant permission set
                    if access_mode in cur_perms["Not Inherited"][user_name]:
                        # If the applies to settings are different, use the new one
                        if (
                            not cur_perms["Not Inherited"][user_name][access_mode][
                                "applies to"
                            ]
                            == applies_to_text
                        ):
                            changes.setdefault(user, {})
                            changes[user]["applies_to"] = applies_to

    if changes:
        ret["changes"].setdefault(perms_label, {})
        for user in changes:
            user_name = get_name(principal=user)

            if test_mode is True:
                ret["changes"][perms_label].setdefault(user, {})
                ret["changes"][perms_label][user] = changes[user]
            else:
                if not test_mode:
                    try:
                        set_permissions(
                            obj_name=obj_name,
                            principal=user_name,
                            permissions=changes[user]["permissions"],
                            access_mode=access_mode,
                            applies_to=changes[user].get("applies_to"),
                            obj_type=obj_type,
                        )
                        ret["changes"].setdefault(perms_label, {}).setdefault(user, {})
                        ret["changes"][perms_label][user] = changes[user]
                    except CommandExecutionError as exc:
                        ret["result"] = False
                        ret["comment"].append(
                            'Failed to change {} permissions for "{}" to {}\n'
                            "Error: {}".format(
                                access_mode, user, changes[user], exc.strerror
                            )
                        )

    return ret


def check_perms(
    obj_name,
    obj_type="file",
    ret=None,
    owner=None,
    grant_perms=None,
    deny_perms=None,
    inheritance=True,
    reset=False,
    test_mode=False,
):
    """
    Check owner and permissions for the passed directory. This function checks
    the permissions and sets them, returning the changes made.

    .. versionadded:: 2019.2.0

    Args:

        obj_name (str):
            The name or full path to the object

        obj_type (Optional[str]):
            The type of object for which to check permissions. Default is 'file'

        ret (dict):
            A dictionary to append changes to and return. If not passed, will
            create a new dictionary to return.

        owner (str):
            The owner to set for the directory.

        grant_perms (dict):
            A dictionary containing the user/group and the basic permissions to
            check/grant, ie: ``{'user': {'perms': 'basic_permission'}}``.
            Default is ``None``.

        deny_perms (dict):
            A dictionary containing the user/group and permissions to
            check/deny. Default is ``None``.

        inheritance (bool):
            ``True`` will enable inheritance from the parent object. ``False``
            will disable inheritance. Default is ``True``.

        reset (bool):
            ``True`` will clear the DACL and set only the permissions defined
             in ``grant_perms`` and ``deny_perms``. ``False`` append permissions
             to the existing DACL. Default is ``False``. This does NOT affect
            inherited permissions.

        test_mode (bool):
            ``True`` will only return the changes that would be made. ``False``
            will make the changes as well as return the changes that would be
            made.

    Returns:
        dict: A dictionary of changes that have been made

    Usage:

    .. code-block:: bash

        # You have to use __utils__ in order for __opts__ to be available

        # To see changes to ``C:\\Temp`` if the 'Users' group is given 'read & execute' permissions.
        __utils__['dacl.check_perms'](obj_name='C:\\Temp',
                                      obj_type='file',
                                      owner='Administrators',
                                      grant_perms={
                                          'Users': {
                                              'perms': 'read_execute'
                                          }
                                      })

        # Specify advanced attributes with a list
        __utils__['dacl.check_perms'](obj_name='C:\\Temp',
                                      obj_type='file',
                                      owner='Administrators',
                                      grant_perms={
                                          'jsnuffy': {
                                              'perms': [
                                                  'read_attributes',
                                                  'read_ea'
                                              ],
                                              'applies_to': 'files_only'
                                          }
                                      })
    """
    # Validate obj_type
    if obj_type.lower() not in flags().obj_type:
        raise SaltInvocationError(f'Invalid "obj_type" passed: {obj_type}')

    obj_type = obj_type.lower()

    if not ret:
        ret = {"name": obj_name, "changes": {}, "comment": [], "result": True}
        orig_comment = ""
    else:
        orig_comment = ret["comment"]
        ret["comment"] = []

    # Check owner
    if owner:
        owner = get_name(principal=owner)
        current_owner = get_owner(obj_name=obj_name, obj_type=obj_type)
        if owner != current_owner:
            if test_mode is True:
                ret["changes"]["owner"] = owner
            else:
                try:
                    set_owner(obj_name=obj_name, principal=owner, obj_type=obj_type)
                    # grant_perms.setdefault(owner, {})
                    log.debug("Owner set to %s", owner)
                    ret["changes"]["owner"] = owner
                except CommandExecutionError:
                    ret["result"] = False
                    ret["comment"].append(f'Failed to change owner to "{owner}"')

    # Check inheritance
    if inheritance is not None:
        if not inheritance == get_inheritance(obj_name=obj_name, obj_type=obj_type):
            if test_mode is True:
                ret["changes"]["inheritance"] = inheritance
            else:
                try:
                    set_inheritance(
                        obj_name=obj_name, enabled=inheritance, obj_type=obj_type
                    )
                    log.debug(
                        "%s inheritance", "Enabling" if inheritance else "Disabling"
                    )
                    ret["changes"]["inheritance"] = inheritance
                except CommandExecutionError:
                    ret["result"] = False
                    ret["comment"].append(
                        'Failed to set inheritance for "{}" to {}'.format(
                            obj_name, inheritance
                        )
                    )

    # Check reset
    # If reset=True, which users will be removed as a result
    if reset:
        log.debug("Resetting permissions for %s", obj_name)
        cur_perms = get_permissions(obj_name=obj_name, obj_type=obj_type)
        for user_name in cur_perms["Not Inherited"]:
            # case-insensitive dictionary search
            if user_name not in {get_name(k) for k in (grant_perms or {})}:
                if "grant" in cur_perms["Not Inherited"][user_name]:
                    ret["changes"].setdefault("remove_perms", {})
                    if test_mode is True:
                        ret["changes"]["remove_perms"].update(
                            {user_name: cur_perms["Not Inherited"][user_name]}
                        )
                    else:
                        rm_permissions(
                            obj_name=obj_name,
                            principal=user_name,
                            ace_type="grant",
                            obj_type=obj_type,
                        )
                        ret["changes"]["remove_perms"].update(
                            {user_name: cur_perms["Not Inherited"][user_name]}
                        )
            # case-insensitive dictionary search
            if user_name not in {get_name(k) for k in (deny_perms or {})}:
                if "deny" in cur_perms["Not Inherited"][user_name]:
                    ret["changes"].setdefault("remove_perms", {})
                    if test_mode is True:
                        ret["changes"]["remove_perms"].update(
                            {user_name: cur_perms["Not Inherited"][user_name]}
                        )
                    else:
                        rm_permissions(
                            obj_name=obj_name,
                            principal=user_name,
                            ace_type="deny",
                            obj_type=obj_type,
                        )
                        ret["changes"]["remove_perms"].update(
                            {user_name: cur_perms["Not Inherited"][user_name]}
                        )

    # Check permissions
    log.debug("Getting current permissions for %s", obj_name)

    # Verify Deny Permissions
    if deny_perms is not None:
        ret = _check_perms(
            obj_name=obj_name,
            obj_type=obj_type,
            new_perms=deny_perms,
            access_mode="deny",
            ret=ret,
            test_mode=test_mode,
        )

    # Verify Grant Permissions
    if grant_perms is not None:
        ret = _check_perms(
            obj_name=obj_name,
            obj_type=obj_type,
            new_perms=grant_perms,
            access_mode="grant",
            ret=ret,
            test_mode=test_mode,
        )

    # Clean up after itself if reset is True
    # This is needed because currently adding a permission will also add all
    # Inherited Permissions as Not Inherited permissions. These will not be
    # added to the Changes dict, as that is handled above
    if reset and not test_mode:
        log.debug("Resetting permissions for %s", obj_name)
        cur_perms = get_permissions(obj_name=obj_name, obj_type=obj_type)
        for user_name in cur_perms["Not Inherited"]:
            # case-insensitive dictionary search
            if user_name not in {get_name(k) for k in (grant_perms or {})}:
                if "grant" in cur_perms["Not Inherited"][user_name]:
                    rm_permissions(
                        obj_name=obj_name,
                        principal=user_name,
                        ace_type="grant",
                        obj_type=obj_type,
                    )
            # case-insensitive dictionary search
            if user_name not in {get_name(k) for k in (deny_perms or {})}:
                if "deny" in cur_perms["Not Inherited"][user_name]:
                    rm_permissions(
                        obj_name=obj_name,
                        principal=user_name,
                        ace_type="deny",
                        obj_type=obj_type,
                    )

    # Re-add the Original Comment if defined
    if isinstance(orig_comment, str):
        if orig_comment:
            ret["comment"].insert(0, orig_comment)
    else:
        if orig_comment:
            ret["comment"] = orig_comment.extend(ret["comment"])

    ret["comment"] = "\n".join(ret["comment"])

    # Set result for test = True
    if test_mode and ret["changes"]:
        ret["result"] = None

    return ret


def _set_perms(obj_dacl, obj_type, new_perms, cur_perms, access_mode):
    obj_type = obj_type.lower()
    ret = {}
    for user in new_perms:
        # Check that user exists:
        user_name = get_name(user)
        # We want to skip unmapped usernames
        if user_name == get_sid_string(user):
            continue

        # Get applies_to
        applies_to = None
        # Propagation only applies to file and registry object types
        if obj_type in ["file", "registry", "registry32"]:
            if "applies_to" not in new_perms[user]:
                # Get current "applies to" settings from the object
                if (
                    user_name in cur_perms["Not Inherited"]
                    and "deny" in cur_perms["Not Inherited"][user_name]
                ):
                    for flag in flags().ace_prop[obj_type]:
                        if (
                            flags().ace_prop[obj_type][flag]
                            == cur_perms["Not Inherited"][user_name]["deny"][
                                "applies to"
                            ]
                        ):
                            at_flag = flag
                            for flag1 in flags().ace_prop[obj_type]:
                                if flags().ace_prop[obj_type][flag1] == at_flag:
                                    applies_to = flag1
                if not applies_to:
                    # Propagation only applies to file and registry object types
                    if obj_type == "file":
                        applies_to = "this_folder_subfolders_files"
                    elif "registry" in obj_type:
                        applies_to = "this_key_subkeys"
            else:
                applies_to = new_perms[user]["applies_to"]

        # Set permissions
        if obj_dacl.add_ace(
            principal=user,
            access_mode=access_mode,
            permissions=new_perms[user]["perms"],
            applies_to=applies_to,
        ):
            ret[user] = new_perms[user]

    return ret


def set_perms(
    obj_name,
    obj_type="file",
    grant_perms=None,
    deny_perms=None,
    inheritance=True,
    reset=False,
):
    """
    Set permissions for the given path

    .. versionadded:: 2019.2.0

    Args:

        obj_name (str):
            The name or full path to the object

        obj_type (Optional[str]):
            The type of object for which to check permissions. Default is 'file'

        grant_perms (dict):
            A dictionary containing the user/group and the basic permissions to
            grant, ie: ``{'user': {'perms': 'basic_permission'}}``. You can also
            set the ``applies_to`` setting here. The default for ``applise_to``
            is ``this_folder_subfolders_files``. Specify another ``applies_to``
            setting like this:

            .. code-block:: yaml

                {'user': {'perms': 'full_control', 'applies_to': 'this_folder'}}

            To set advanced permissions use a list for the ``perms`` parameter,
            ie:

            .. code-block:: yaml

                {'user': {'perms': ['read_attributes', 'read_ea'], 'applies_to': 'this_folder'}}

            To see a list of available attributes and applies to settings see
            the documentation for salt.utils.win_dacl.

            A value of ``None`` will make no changes to the ``grant`` portion of
            the DACL. Default is ``None``.

        deny_perms (dict):
            A dictionary containing the user/group and permissions to deny along
            with the ``applies_to`` setting. Use the same format used for the
            ``grant_perms`` parameter. Remember, deny permissions supersede
            grant permissions.

            A value of ``None`` will make no changes to the ``deny`` portion of
            the DACL. Default is ``None``.

        inheritance (bool):
            If ``True`` the object will inherit permissions from the parent, if
            ``False``, inheritance will be disabled. Inheritance setting will
            not apply to parent directories if they must be created. Default is
            ``False``.

        reset (bool):
            If ``True`` the existing DCL will be cleared and replaced with the
            settings defined in this function. If ``False``, new entries will be
            appended to the existing DACL. Default is ``False``.

    Returns:
        bool: True if successful

    Raises:
        CommandExecutionError: If unsuccessful

    Usage:

    .. code-block:: bash

        import salt.utils.win_dacl

        # To grant the 'Users' group 'read & execute' permissions.
        salt.utils.win_dacl.set_perms(obj_name='C:\\Temp',
                                      obj_type='file',
                                      grant_perms={
                                          'Users': {
                                              'perms': 'read_execute'
                                          }
                                      })

        # Specify advanced attributes with a list
        salt.utils.win_dacl.set_perms(obj_name='C:\\Temp',
                                      obj_type='file',
                                      grant_perms={
                                          'jsnuffy': {
                                              'perms': [
                                                  'read_attributes',
                                                  'read_ea'
                                              ],
                                              'applies_to': 'this_folder_only'
                                          }
                                      }"
    """
    ret = {}

    if reset:
        # Get an empty DACL
        obj_dacl = dacl(obj_type=obj_type)

        # Get an empty perms dict
        cur_perms = {"Inherited": {}, "Not Inherited": {}}

    else:
        # Get the DACL for the directory
        obj_dacl = dacl(obj_name, obj_type=obj_type)

        # Get current file/folder permissions
        cur_perms = get_permissions(obj_name=obj_name, obj_type=obj_type)

    # Set 'deny' perms if any
    if deny_perms is not None:
        ret["deny"] = _set_perms(
            obj_dacl=obj_dacl,
            obj_type=obj_type,
            new_perms=deny_perms,
            cur_perms=cur_perms,
            access_mode="deny",
        )

    # Set 'grant' perms if any
    if grant_perms is not None:
        ret["grant"] = _set_perms(
            obj_dacl=obj_dacl,
            obj_type=obj_type,
            new_perms=grant_perms,
            cur_perms=cur_perms,
            access_mode="grant",
        )

    # Order the ACL
    obj_dacl.order_acl()

    # Save the DACL, setting the inheritance
    # you have to invert inheritance because dacl.save is looking for
    # protected. protected True means Inherited False...

    if obj_dacl.save(obj_name, not inheritance):
        return ret

    return {}
